#ifndef _BT_ASSET_H
#define _BT_ASSET_H

#include <ck_cc.h>
#include <ck_hs.h>
#include <limits.h>
#include <stdbool.h>
#include <sys/queue.h>

#include "gvc.h"
#include "lut.h"
#include "misc.h"
#include "tc.h"

#define BT_ASSET_SUFFIX_MAX 16

#define BT_ASSET_MAGIC_VALUE	"BTAF"
#define BT_ASSET_VERSION 0x01

#define BT_ASSET_TC		(1UL << 0)
#define BT_ASSET_GVC		(1UL << 1)

const char *bt_asset_section_name(unsigned long code);

enum bt_asset_file_mode {
	BT_ASSET_MODE_READ_ONLY,
	BT_ASSET_MODE_READ_WRITE
};

typedef size_t bt_asset_section_size_fn_t(void *);
typedef bool bt_asset_section_write_fn_t(unsigned char *, void *);

/*
 * Describes a section to be written to an asset file.
 * Clients wanting to write new sections must provide a callback to indicate
 * how much space they will need before actually writing any data.
 * The closure variable will be passed as an argument to the size and write
 * callbacks.
 */
struct bt_asset_section_context {
	unsigned long long code;
	unsigned int version;
	bt_asset_section_size_fn_t *size;
	bt_asset_section_write_fn_t *write;
	void *closure;
};

struct bt_bf;
struct bt_gv_query;
struct bt_process;

BT_ARRAY(suffix);

/* Provides all parameters needed to determine usable asset file paths. */
struct bt_asset_key {

	/* The binary file to which this asset data applies. */
	struct bt_bf *bf;

	/*
	 * A list of file extensions to use.  Each of these will be tried in
	 * order for each potential path type before moving on to the next path.
	 */
	BT_ARRAY(suffix) *suffixes;

	/*
	 * Users can (optionally) manually specify a directory for storing
	 * asset files.  This path must be absolute.  If this path is non-empty, 
	 * then it will take precedence over any other paths that would
	 * normally be checked.
	 */
	char manual_dir[PATH_MAX];

	/*
	 * Users can also specify a complete path for the asset file itself.
	 * This takes precedence over any other options.  This path must also
	 * be absolute.
	 */
	char manual_path[PATH_MAX];

	enum bt_asset_file_mode mode;
};
typedef struct bt_asset_key bt_asset_key_t;

bool bt_asset_key_init(bt_asset_key_t *);
void bt_asset_key_destroy(bt_asset_key_t *);

bool bt_asset_key_suffix_add(bt_asset_key_t *, const char *);
bool bt_asset_key_manual_dir_set(bt_asset_key_t *, const char *);
bool bt_asset_key_manual_path_set(bt_asset_key_t *, const char *);
void bt_asset_key_bf_set(bt_asset_key_t *, struct bt_bf *);
void bt_asset_key_mode_set(bt_asset_key_t *, enum bt_asset_file_mode);

/* Represents an individual file on disk containing asset data. */
struct bt_asset_file {
	char path[PATH_MAX];
	int fd;
	unsigned char *mmap;
	size_t length;
	enum bt_asset_file_mode mode;

	struct {
		struct bt_asset_section_context *contexts;
		unsigned int count;
		unsigned int capacity;
	} write;
};
typedef struct bt_asset_file bt_asset_file_t;

bool bt_asset_file_section_add(bt_asset_file_t *,
    struct bt_asset_section_context *);

struct bt_am_iterator;
struct bt_am_file_iterator;

bool bt_asset_section_remove(struct bt_am_iterator *);
bool bt_asset_file_section_remove(struct bt_am_file_iterator *);

/*
 * Commit all sections that have been added.
 * This must be called before closing a file or else the changes will
 * not be written to disk.
 */
bool bt_asset_file_write(bt_asset_file_t *);
void bt_asset_file_close(bt_asset_file_t *);

BT_ARRAY(gvc);

/*
 * This is used to represent both objects that have been loaded into a cache
 * as well as objects that users have requested that are still pending loading.
 */
struct bt_asset_gv_record {
	const char *object;
	const char **cus;
	unsigned int n_cu;
	bt_gvc_t *cache;
	LIST_ENTRY(bt_asset_gv_record) linkage;
};
typedef struct bt_asset_gv_record bt_asset_gv_record_t;

/*
 * Contains all asset data for a binary file.
 * Data may come from multiple underlying files on disk,
 * but this structure provides a unified view of that data.
 * Clients should work with this type instead of directly
 * with asset files.
 */
struct bt_asset {
	struct bt_process *process;

	/* Indicates which asset types have been loaded. */
	unsigned long loaded;

	struct {
		bt_asset_file_t *files;
		unsigned int n_file;
		unsigned int file_capacity;
	} files;

	struct {
		/*
		 * Tracks which objects have already been loaded from on-disk
		 * caches so we can avoid duplicating work during runtime
		 * loading.  Also indicates which cache has records for each
		 * object.
		 */
		ck_hs_t loaded;

		/*
		 * Lists objects which are pending runtime loading.
		 */
		LIST_HEAD(, bt_asset_gv_record) pending;

		BT_ARRAY(gvc) *caches;
	} globals;
	bt_tc_t type_cache;
};
typedef struct bt_asset bt_asset_t;

bool bt_asset_init(bt_asset_t *);
void bt_asset_close(bt_asset_t *);

struct bt_config_load;

bool bt_asset_gv_register(bt_asset_t *, const struct bt_config_load *);
void bt_asset_process(bt_asset_t *, struct bt_process *);

struct bt_config;
bool bt_asset_monolithic_init(struct bt_config *);
bt_asset_t *bt_asset_monolithic_instance(void);

/*
 * Searches all paths indicated by the given key and loads data from
 * any asset files found at those paths.  If write mode is indicated,
 * only the first valid path found will be used, and all sections
 * will be written to that file.
 *
 * If a client wants to write multiple sections to different asset files,
 * they should use the asset file interface directly.
 */
bool bt_asset_load(bt_asset_t *, unsigned long, struct bt_asset_key *);

static inline bool
bt_asset_loaded(bt_asset_t *asset, unsigned long code)
{

	return asset->loaded & code;
}

bt_tc_t *bt_asset_type_cache(bt_asset_t *);

struct bt_asset_gv_iterator {
	bt_asset_t *asset;
	const struct bt_gv_query *query;

	enum {
		BT_ASSET_GV_CACHE_ALL,
		BT_ASSET_GV_CACHE_SINGLE,
		BT_ASSET_GV_CACHE_NONE
	} cache_type;

	unsigned int ix;
	bool match;
	enum {
		BT_ASSET_GV_STATE_INIT,
		BT_ASSET_GV_STATE_DISK,
		BT_ASSET_GV_STATE_RUNTIME
	} state;

	bt_gvc_iterator_t gvc_iter;
};
typedef struct bt_asset_gv_iterator bt_asset_gv_iterator_t;

enum bt_asset_gv_iterator_result {
	BT_ASSET_GV_ITERATOR_OK,	/* Result was found */
	BT_ASSET_GV_ITERATOR_ERROR,	/* An error occurred */
	BT_ASSET_GV_ITERATOR_DONE	/* No results remain */
};

bool bt_asset_gv_iterator_init(bt_asset_gv_iterator_t *, bt_asset_t *,
    const struct bt_gv_query *);
enum bt_asset_gv_iterator_result bt_asset_gv_iterator_next(
    bt_asset_gv_iterator_t *, struct bt_gvc_iterator_entry *);
void bt_asset_gv_iterator_deinit(bt_asset_gv_iterator_t *);

struct bt_asset_section_header;

struct bt_am_file_iterator {
	bt_asset_file_t *file;
	struct bt_asset_section_header *header;
	unsigned long codes;
	bool section_removed;
	bool started;
};

struct bt_am_iterator {
	bt_asset_t *asset;
	unsigned int file_index;
	struct bt_am_file_iterator file_iter;
	unsigned long codes;
};

/* Stores data for each section found by an iterator. */
struct bt_asset_section_data {
	unsigned long code;
	unsigned int version;
	size_t length;
	void *data;
};

bool bt_am_iterator_init(struct bt_am_iterator *, bt_asset_t *,
    unsigned long);
bool bt_am_iterator_next(struct bt_am_iterator *,
    struct bt_asset_section_data *);

void bt_am_file_iterator_init(struct bt_am_file_iterator *,
    struct bt_asset_file *, unsigned long);
bool bt_am_file_iterator_next(struct bt_am_file_iterator *,
    struct bt_asset_section_data *);

struct bt_asset_config {
	const char *map_path;
};
typedef struct bt_asset_config bt_asset_config_t;

#define BT_ASSET_CONFIG_INITIALIZER { .map_path = NULL }

static inline CK_CC_UNUSED void
bt_asset_config_map_path_set(bt_asset_config_t *config, const char *path)
{

	config->map_path = path;
	return;
}

static inline CK_CC_UNUSED const char *
bt_asset_config_map_path(const bt_asset_config_t *config)
{

	return config->map_path;
}

#endif /* _BT_ASSET_H */
