#ifndef _BT_BF_H
#define _BT_BF_H

#include <ck_ht.h>
#include <limits.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/queue.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "bt_unwind.h"
#include "dwarf.h"
#include "lut.h"
#include "misc.h"

enum bt_bf_format {
	BT_BF_FORMAT_UNKNOWN = 0,
	BT_BF_FORMAT_ELF,
	BT_BF_FORMAT_MACHO
};

enum bt_bf_type {
	BT_BF_TYPE_EXEC = 0,
	BT_BF_TYPE_DYN,
	BT_BF_TYPE_REL,
	BT_BF_TYPE_CORE,
	BT_BF_TYPE_OTHER
};

enum bt_bf_class {
	BT_BF_CLASS_NATIVE = 0,
	BT_BF_CLASS_32,
	BT_BF_CLASS_64
};

struct bt_bf;
typedef struct bt_bf bt_bf_t;

struct bt_bf_section {
	bt_bf_t *bf;
	const char *name;
	void *md;
};
typedef struct bt_bf_section bt_bf_section_t;

enum bt_bf_section_compression {
	BT_BF_SECTION_C_NONE = 0,
	BT_BF_SECTION_C_ZLIB,
	BT_BF_SECTION_C_UNKNOWN
};

struct bt_bf_section_chdr {
	enum bt_bf_section_compression compression;
	unsigned int data_offset;
	uint64_t uncompressed_size;
	uint64_t addralign;
};
typedef struct bt_bf_section_chdr bt_bf_section_chdr_t;

/*
 * ELF, PE, and Mach-O all have sections, and this struct
 * contains the data that is common between them all.
 * Nothing that is format specific.
 */
struct bt_bf_section_stat {
	uintptr_t addr;
	uint64_t offset;
	size_t size;
};

static inline const char *
bt_bf_section_name(const struct bt_bf_section *section)
{

	return section->name;
}

#define BT_BF_SYMBOL_F_TLS		(1UL << 0)

struct bt_bf_symbol {
	bt_bf_t *bf;
	bt_bf_section_t *section;
	const char *demangle;
	const char *name;
	unsigned long flags;

	/* In the case of TLS variables, this represents the TLS offset */
	unsigned long long address;

	unsigned long long size;
	void *md;
};
typedef struct bt_bf_symbol bt_bf_symbol_t;

const char *bt_bf_symbol_demangle(bt_bf_symbol_t *);

static inline unsigned long long
bt_bf_symbol_size(const struct bt_bf_symbol *symbol)
{

	return symbol->size;
}

static inline unsigned long long
bt_bf_symbol_address(const struct bt_bf_symbol *symbol)
{

	return symbol->address;
}

static inline const char *
bt_bf_symbol_name(const struct bt_bf_symbol *symbol)
{

	return symbol->name;
}

static inline void *
bt_bf_symbol_md(const struct bt_bf_symbol *symbol)
{

	return symbol->md;
}

struct bt_bf_symbol_cache {
	/* Name to bt_bf_symbol index. */
	ck_ht_t name;

	/* Address to symbol cache. */
	struct {
		struct bt_bf_symbol **symbol;
		size_t length;
		size_t capacity;
	} address;
};
typedef struct bt_bf_symbol_cache bt_bf_symbol_cache_t;

#define BT_BF_BUILD_ID_LENGTH 23
#define BT_BF_BUILD_ID_STRING_LENGTH (BT_BF_BUILD_ID_LENGTH * 2 + 1)

struct bt_bf_build_id {
	unsigned char id[BT_BF_BUILD_ID_LENGTH];
	unsigned char length;
} CK_CC_PACKED;
typedef struct bt_bf_build_id bt_bf_build_id_t;

void bt_bf_build_id_print(char *, bt_bf_build_id_t *);
bool bt_bf_build_id_equal(const bt_bf_build_id_t *, const bt_bf_build_id_t *);


#define BT_BF_AUXILIARY		(1U << 0)
#define BT_BF_ASSETS		(1U << 1)
/* Do not apply any relocations specified in this file */
#define BT_BF_NO_RELOCATE	(1U << 2)

/* Do not attempt to respect sysroot. */
#define BT_BF_ABSOLUTE		(1U << 3)

struct bt_bf {
	unsigned int state;
	unsigned long flags;
	enum bt_bf_type type;
	enum bt_bf_class class;
	bt_bf_t *pair;
	unsigned int ref;
	char *basename;

	/*
	 * Some files specify an alternative basename to use for finding
	 * auxiliary debug files.
	 */
	const char *debug_link;

	char path[PATH_MAX];
	struct stat st;
	bt_bf_build_id_t build;

	/* Line DWARF state machine. */
	bt_dwarf_lvm_t lvm;

	/* DWARF type state machine. */
	bt_dwarf_t dwarf;

	/* Unwinding information for this object */
	bt_unwind_file_t unwind;

	union {
		uint32_t vaddr_32;
		uint64_t vaddr_64;
	} base;

	unsigned long long map;
	unsigned long long link_map;
	unsigned long long tls_base;
	unsigned long long entry;
	unsigned long long phdr_base;

	int fd;
	void *md;
	bt_bf_symbol_cache_t symbol_cache;
	struct bt_bf_adapter *adapter;

	ck_ht_t section_cache;
	bt_lut_t sections;
};

/*
 * This flag is pertinent to Mach-O binaries which may be
 * multiarch (Supporting two different architectures).
 */
#define BT_BF_HEADER_MD_F_MULTIARCH	(1U << 0)

struct bt_bf_header_md {
	enum bt_bf_class class; /* Supported architecture */
	unsigned long long entry; /* Entry point. */
	unsigned int flag;
};

/*
 * These flag values tell if the text and data segment were
 * found, or if the text and data are merged into one segment.
 */
#define BT_BF_SEGMENT_MD_F_TEXT	(1U << 0)
#define BT_BF_SEGMENT_MD_F_DATA	(1U << 1)
#define BT_BF_SEGMENT_MD_F_MISC	(1U << 2)

struct bt_bf_segment_md_entry {
	unsigned long long addr;
	unsigned long long offset;
	size_t fsize;
	size_t msize;
	unsigned int perm;
	unsigned int flag;
};

/*
 * This struct contains general segment information for a
 * binary format.
 */
struct bt_bf_segment_md {
	struct bt_bf_segment_md_entry *load;
	size_t load_count;
};

/*
 * Resolves line associated with address.
 */
bool bt_bf_resolve_pc(bt_bf_t *, unsigned long, struct bt_dwarf_line *,
    unsigned long);

/*
 * Copies all map address data from the second file to the first.
 */
void bt_bf_copy_mapping(bt_bf_t *, const bt_bf_t *);

unsigned long long bt_bf_relocate(const bt_bf_t *);

/*
 * Resolves frame state associated with address.
 */
struct bt_unwind_cursor;
struct bt_frame;
bool bt_bf_resolve_frame(bt_bf_t *, struct bt_frame *, unsigned long);

static inline bool
bt_bf_type(const struct bt_bf *bf, enum bt_bf_type type)
{

	return bf->type == type;
}

static inline bool
bt_bf_class(struct bt_bf *bf, enum bt_bf_class class)
{

	return bf->class == class;
}

static inline bool
bt_bf_relocate_enabled(const struct bt_bf *bf)
{

	return (bf->flags & BT_BF_NO_RELOCATE) == 0;
}

void bt_bf_build_set(bt_bf_t *, const unsigned char *, size_t);

/*
 * Associates a section by name.
 */
bool bt_bf_section_assoc_name(bt_bf_t *, bt_bf_section_t *, const char *);

/*
 * Set type of binary file.
 */
void bt_bf_type_set(bt_bf_t *, enum bt_bf_type);

/*
 * Set class of binary file.
 */
void bt_bf_class_set(bt_bf_t *, enum bt_bf_class);

/*
 * Get class of binary file.  This will always return a specific class, i.e.
 * 32 or 64, as opposed to CLASS_NATIVE.
 */
enum bt_bf_class bt_bf_class_get(const bt_bf_t *);

/*
 * Get the debug class of binary file.
 */
enum bt_dwarf_debug_class bt_bf_debug_class_get(const bt_bf_t *);

/*
 * Set address of linker map entry for binary file.
 */
void bt_bf_link_map_set(bt_bf_t *, unsigned long long);

/*
 * Returns a pointer to the debug information for this file, initializing it if
 * necessary.
 * A return value of NULL inidicates that debug initialization failed.
 */
bt_dwarf_t *bt_bf_dwarf(bt_bf_t *);

/*
 * Returns a pointer to the unwinder for this file, initializing it if
 * necessary.
 * A return value of NULL inidicates that unwinder initialization failed.
 */
bt_unwind_file_t *bt_bf_unwind(bt_bf_t *);

/*
 * Capture the entry point of an ET_EXEC or ET_DYN ELF file.
 */
void bt_bf_entry_set(bt_bf_t *, unsigned long long);

/*
 * Return the recorded entry point value.
 */
static inline unsigned long long
bt_bf_entry_get(const struct bt_bf *bf)
{

	return bf->entry;
}

static inline const char *
bt_bf_basename(const struct bt_bf *bf)
{

	return bf->basename;
}

static inline const char *
bt_bf_debug_link(const struct bt_bf *bf)
{

	return bf->debug_link;
}

static inline const unsigned char *
bt_bf_build_id(struct bt_bf *bf, unsigned char *length)
{

	*length = bf->build.length;
	if (bf->build.length > 0)
		return bf->build.id;

	return NULL;
}

struct bt_bf_mmap {
	/*
	 * Metadata for data to be exposed to user.  This can differ from the
	 * mmap metadata because we need to ensure that mmapped regions are
	 * page aligned.
	 *
	 * For example, suppose that we want to map a file segment starting at
	 * file offset 0x123456.  If we want to use the mmap interface, we
	 * would have to map from that file starting at offset 0x123000
	 * (assuming 4K pages) and adjust the mapping size accordingly.
	 *
	 * We maintain two sets of metadata to account for this.  The first set
	 * represents the logical state of the data as requested by the user.
	 * We use this data for all user-facing interfaces.  We use the second
	 * set of data for interacting with mmap or other backing interfaces.
	 *
	 * Consider a request to map a segment of size 0x2222 from file offset
	 * 0x123456 on a system with 4K pages.  Because this segment begins
	 * 0x456 bytes past a page boundary, we would instead map from file
	 * offset 0x123000 and increase the mapping size to
	 * 0x2222 + 0x456 = 0x2678.  If the resulting mapped data had address
	 * 0xabc000, then the first set of metadata would have address
	 * 0xabc456 and size 0x2222, and the second set would have address
	 * 0xabc000 and size 0x2678.
	 */
	struct {
		const void *address;	/* Start of logical data */
		size_t size;		/* Size of logical data */
	} data;

	/*
	 * Metadata for mmap'd region.  This data should only be used for
	 * interacting with mmap and related interfaces.
	 */
	struct {
		void *region;		/* Beginning of mapping. */
		size_t size;		/* Size of mapping */
	} mmap;

	unsigned long flags;
	int prot;
};
typedef struct bt_bf_mmap bt_bf_mmap_t;

/*
 * mmap descriptor requires initialization before any use.
 */
static inline void
bt_bf_mmap_init(struct bt_bf_mmap *m)
{

	memset(m, 0, sizeof *m);
	return;
}

/*
 * Returns data associated with mmap descriptor.
 */
static inline const void *
bt_bf_mmap_data(struct bt_bf_mmap *m)
{

	return m->data.address;
}

/*
 * Maps a region of memory using the specified mmap descriptor.
 * If the mapped region falls below a page then malloc may be
 * used.
 */
const void *bt_bf_mmap(bt_bf_mmap_t *, bt_bf_t *, size_t, off_t);

/*
 * Copy data into a mapped region
 * Arguments:
 *	offset: Offset into region to copy to
 *	buf: Buffer containing data to copy
 *	bytes: Number of bytes to copy
 */
bool bt_bf_mmap_memcpy(bt_bf_mmap_t *m, off_t offset,
    const void * restrict buf, size_t bytes);

/*
 * Unmaps the region of memory associated with the descriptor.
 */
void bt_bf_munmap(bt_bf_mmap_t *);

static inline void
bt_bf_addref(struct bt_bf *bf)
{

	bf->ref++;
	return;
}

static inline bool
bt_bf_delref(struct bt_bf *bf)
{

	return --bf->ref == 0;
}

/*
 * This currently assumes that maps are processed in order.
 * This will likely have to change for non-Linux where maps
 * entries guaranteed to be in order.
 */
static inline void
bt_bf_map_set(struct bt_bf *bf, unsigned long long map)
{

	if (bf->pair != NULL)
		bf->pair->map = map;

	bf->map = map;
	return;
}

static inline unsigned long long
bt_bf_map(const struct bt_bf *bf)
{

	return bf->map;
}

static inline unsigned long long
bt_bf_base(const struct bt_bf *bf)
{

	return bf->base.vaddr_64;
}

static inline unsigned long long *
bt_bf_link_map(struct bt_bf *bf)
{

	return &bf->link_map;
}

static inline unsigned long long
bt_bf_tls_base(const struct bt_bf *bf)
{

	return bf->tls_base;
}

bool bt_bf_symbol_cache(bt_bf_t *, bt_bf_symbol_t *, bt_bf_section_t *,
    void *, size_t);

/*
 * Returns the symbol of the specified name or NULL if one doesn't
 * exist.
 */
bt_bf_symbol_t *bt_bf_symbol_name_get(bt_bf_t *, const char *);

static inline void
bt_bf_symbol_init(bt_bf_symbol_t *symbol, const char *n, const char *d,
    uint64_t a, size_t s, bool t, void *m)
{

	symbol->name = n;
	symbol->demangle = d;
	symbol->address = a;
	symbol->size = s;
	symbol->flags = 0;
	symbol->md = m;

	if (t == true)
		symbol->flags |= BT_BF_SYMBOL_F_TLS;

	return;
}

static inline bool
bt_bf_symbol_tls(const bt_bf_symbol_t *symbol)
{

	return symbol->flags & BT_BF_SYMBOL_F_TLS;
}

/* Provides a size hint to the symbol cache. */
bool bt_bf_symbol_cache_hint(bt_bf_t *, size_t);

enum bt_bf_iterator_result {
	BT_BF_ITER_OK,
	BT_BF_ITER_DONE,
	BT_BF_ITER_ERROR
};

struct bt_bf_map_iterator {
	bt_bf_t *bf;
	unsigned int i;
};
typedef struct bt_bf_map_iterator bt_bf_map_iterator_t;

struct bt_bf_map {
	unsigned long long start;
	unsigned long long end;
	unsigned long long offset;
	size_t data_size;
	unsigned int p_flags;
};

struct bt_bf_note_iterator {
	bt_bf_t *bf;
	void *md;
};
typedef struct bt_bf_note_iterator bt_bf_note_iterator_t;

struct bt_bf_note {
	unsigned long type;
	const char *name;
	const char *data;
	size_t data_size;
};

struct bt_so_resolver;
struct bt_process;

struct bt_bf_adapter {
	/*
	 * Determines whether the file may be opened with the specified
	 * adapter and loads initial state. If the 3rd argument is passed
	 * as non-NULL, then a stat of the binary header information will
	 * be attempted, and the success of this will be stored in the 4th
	 * argument.
	 */
	bool (*open)(bt_bf_t *, void **, struct bt_bf_header_md *, bool *);

	/*
	 * Reads in meta-data.
	 */
	bool (*read)(bt_bf_t *);

	/*
	 * Returns an auxiliary file with debug information if one is
	 * specified in the file provided.  Returns NULL otherwise.
	 */
	bt_bf_t *(*pair)(bt_bf_t *);

	/*
	 * Returns pointer to contents of a section and indicates
	 * size of section. Returns NULL on failure.
	 */
	const void *(*read_section)(bt_bf_t *, bt_bf_section_t *, size_t *,
	    bt_bf_section_chdr_t *);

	/*
	 * Fills in a bt_bf_section_stat struct with section meta-data
	 */
	void (*section_stat)(bt_bf_section_t *, struct bt_bf_section_stat *);

	/*
	 * Fills in bt_bf_segment_md with basic segment meta-data.
	 */
	bool (*segment_stat)(bt_bf_t *, struct bt_bf_segment_md *);

	/*
	 * Fills in bt_bf_header_md with common binary format file
	 * header information: arch, entry, multiarch
	 */
	void (*header_stat)(bt_bf_t *, struct bt_bf_header_md *);

	/*
	 * Prints format-specific object information.
	 */
	void (*print)(FILE *, bt_bf_t *, int);

	/*
	 * Prints format-specific section information.
	 */
	void (*print_section)(FILE *, bt_bf_section_t *, int);

	/*
	 * Prints format-specific symbol information.
	 */
	void (*print_symbol)(FILE *, bt_bf_symbol_t *, int);

	/*
	 * Canonocalizes a symbol table based on symbols which meet the criteria
	 * defined by the callback function.
	 */
	unsigned long (*symtab)(bt_bf_t *, bool (*)(bt_bf_symbol_t *, void *), void *);

	/*
	 * Returns a pointer to a bt_bf_section_t containing the metadata for
	 * the '.text' section.
	 */
	bt_bf_section_t * (*get_text_section)(bt_bf_t *);

	/*
	 * Returns a pointer to the .interp section which contains a path to
	 * the program interpreter (aka dynamic linker)
	 */
	const char * (*get_interpreter_path)(bt_bf_t *);

	/*
	 * Returns a pointer to a NUL-separated string of names of shared
	 * objects on which the file depends.  The caller is responsible for
	 * freeing this string.
	 *
	 * The final name will be terminated with 2 null bytes instead of one.
	 * If the resolver struct pointer is non-NULL, it will be populated
	 * with any relevant data from the file.
	 *
	 * The strings provided are basenames only; the caller is responsible
	 * for resolving the strings to full paths using the resolver struct.
	 */
	char *(*get_shared_objects)(bt_bf_t *, struct bt_so_resolver *);

	/*
	 * Returns the base pointer for thread debugging information for this
	 * file. This is required for TLS support.
	 */
	unsigned long long (*get_debug_base)(bt_bf_t *, struct bt_process *);

	/*
	 * Initializes an iterator to provide all mapped memory segments
	 * from a core file.
	 */
	bool (*map_iterator_init)(bt_bf_map_iterator_t *, bt_bf_t *,
	    unsigned int *);

	/*
	 * Populates the structure in the second argument with information
	 * about the next mapped memory segment from a core file.
	 * The contents of the second argument are only valid if the return
	 * value is BT_BF_MAP_OK.
	 */
	enum bt_bf_iterator_result (*map_iterator_next)(
	    bt_bf_map_iterator_t *, struct bt_bf_map *);

	bool (*note_iterator_init)(bt_bf_note_iterator_t *, bt_bf_t *);

	enum bt_bf_iterator_result (*note_iterator_next)(
	    bt_bf_note_iterator_t *, struct bt_bf_note *);

	void (*note_iterator_deinit)(bt_bf_note_iterator_t *);

	/*
	 * Some formats provide alternative names for auxiliary debug files.
	 * This returns such a name if one exists, or NULL otherwise.
	 */
	const char *(*get_debug_link)(bt_bf_t *);

	/*
	 * Destroys MD-specific data associated with object.
	 */
	void (*destroy)(bt_bf_t *);
	enum bt_bf_format format;
	LIST_ENTRY(bt_bf_adapter) _linkage;
};

/* First check paired file (if one exists) for section */
#define BT_BF_SECTION_PAIR	(1UL << 0)

/*
 * Returns a section with the given name or NULL if it does not
 * exist.
 */
bt_bf_section_t *bt_bf_section_find(bt_bf_t *, const char *, unsigned long);

/*
 * Creates an abstract binary file section. It is possible to associate
 * adapter-specific information with the section through md and md_size
 * arguments.
 */
bool bt_bf_section_set(bt_bf_t *bf, const char *name,
    size_t index, void *md, size_t md_size);

/*
 * Returns a pointer to an abstract section object given the specified
 * index.
 */
bt_bf_section_t *bt_bf_section_get(bt_bf_t *bf, unsigned int index);

/*
 * Returns a pointer to the data contained in a section. Last argument
 * is set to section size.
 */
const void *bt_bf_section_read(bt_bf_t *, bt_bf_section_t *, size_t *,
    struct bt_bf_section_chdr *);

/*
 * Returns line virtual machine associated with a binary object.
 */
static inline bt_dwarf_lvm_t *
bt_bf_lvm(struct bt_bf *bf)
{

	return &bf->lvm;
}

static inline void *
bt_bf_section_md(struct bt_bf_section *section)
{

	return section->md;
}

static inline enum bt_bf_format
bt_bf_format(const struct bt_bf *bf)
{

	return bf->adapter->format;
}

/*
 * Print contents of a section in human-readable format.
 */
void bt_bf_section_print(FILE *, bt_bf_t *, bt_bf_section_t *,
    unsigned long long);

char *bt_bf_shared_objects(bt_bf_t *, struct bt_so_resolver *);

unsigned long long bt_bf_debug_base(bt_bf_t *, struct bt_process *);

/*
 * Initialize an iterator to provide all map entries from a core file.
 * Return the number of program headers through the third paramenter if
 * not a NULL pointer.
 */
bool bt_bf_map_iterator_init(bt_bf_map_iterator_t *, bt_bf_t *,
    unsigned int *);

/*
 * Get data for the next map entry in a core file.  If the return value is
 * BT_BF_ITER_OK, then data for a map entry will be written to the second
 * argument.  Otherwise, the contents of that argument are undefined after
 * the call.  A return value of BT_BF_ITER_ERROR indicates that an error
 * occurred reading map entries.  A return value of BT_BF_ITER_DONE indicates
 * that all map entries have been provided already and is not an error
 * condition.
 */
enum bt_bf_iterator_result bt_bf_map_iterator_next(bt_bf_map_iterator_t *,
    struct bt_bf_map *);

/*
 * Initialize an iterator to provide data for all notes in a core file.
 */
bool bt_bf_note_iterator_init(bt_bf_note_iterator_t *, bt_bf_t *);

/*
 * Get data for the next note in a core file.  The return values and semantics
 * for the second argument are the same as for the map iterator above.
 */
enum bt_bf_iterator_result bt_bf_note_iterator_next(bt_bf_note_iterator_t *,
    struct bt_bf_note *);

/*
 * Destroy any temporary resources created by a note iterator.  This method
 * should be called on an iterator after the caller is done using it, regardless
 * of whether or not an error occurred.
 */
void bt_bf_note_iterator_deinit(bt_bf_note_iterator_t *);

#define BT_BF_FOREACH_SECTION(B, S)				\
	for (unsigned int _bt_i = 0;				\
	    _bt_i < bt_lut_length(&(B)->sections) &&		\
	    ((S) = bt_lut_get(&(B)->sections, _bt_i++),1);)

#define BT_BF_FOREACH_SYMBOL(B, S)				\
	for (unsigned int _bt_i = 0;				\
	    _bt_i < (B)->symbol_cache.address.length &&		\
	    ((S) = (B)->symbol_cache.address.symbol[_bt_i++],1);)

/*
 * Sets binary file to an immutable state suitable for
 * symbol queries.
 */
bool bt_bf_init(void);
void bt_bf_adapter_register(struct bt_bf_adapter *);
size_t bt_bf_pread(bt_bf_t *, void *, size_t, off_t);

/*
 * If auxiliary flag is present, the auxiliary debug
 * files will be attempted to be loaded.
 */
bt_bf_t *bt_bf_open(const char *, unsigned int, mode_t);

/*
 * This is the function which bt_bf_open uses to open a binary.
 * It may also be called directly when the caller would like to stat the
 * binaries header, in this case the last two arguments must be non NULL.
 */
bt_bf_t *bt_bf_open_md(const char *, unsigned int, mode_t,
    struct bt_bf_header_md *, bool *);

/*
 * Foce an auxiliary mapping.
 */
#define BT_BF_AUXILIARY_FORCE	(1U << 0)

void bt_bf_close(bt_bf_t *);
bool bt_bf_read(bt_bf_t *);
void bt_bf_print(FILE *, bt_bf_t *, unsigned long long);

static inline const char *
bt_bf_path(const struct bt_bf *bf)
{

	return bf->path;
}

static inline void *
bt_bf_md(struct bt_bf *bf)
{

	return bf->md;
}

static inline unsigned long long bt_bf_phdr_base(struct bt_bf *bf)
{

	return bf->phdr_base;
};

/* Returns a symbol associated with provided address. */
bt_bf_symbol_t *bt_bf_symbol_address_get(bt_bf_t *, unsigned long);

/*
 * Search for an auxiliary debug file using GNU/LSB semantics.
 * Returns NULL if no auxiliary file is found.
 */
bt_bf_t *bt_bf_pair_gnu(bt_bf_t *, unsigned long);

/*
 * An exposed interface function for elf_adapter->section_stat.
 */
void bt_bf_section_stat(bt_bf_section_t *, struct bt_bf_section_stat *);

/*
 * Function for acquiring segment metdata about a binary file.
 * It essentially returns information about the loadable segments.
 * These are common among all executable formats, and primarily
 * there exists a text and a data segment usually. There are several
 * flags that the caller may check to see if the given load segment
 * is a text segment, data segment, misc segment, or a merged segment.
 * A merged segment is when the text and data segment are merged into
 * a single segment
 */
bool bt_bf_segment_stat(bt_bf_t *, struct bt_bf_segment_md *);

/*
 * Should be called after bt_bf_segment_stat to free any allocated
 * memory.
 */
void bt_bf_segment_stat_deinit(struct bt_bf_segment_md *);

/*
 * Returns a pointer to section_t for the ".text" section
 */
bt_bf_section_t *bt_bf_get_text_section(struct bt_bf *);

/*
 * Returns a pointer to the path found in the .interp section
 * which is the path to the programs dynamic linker.
 */
const char * bt_bf_get_interp_path(struct bt_bf *);

enum bt_bf_ctors_type {
	BT_BF_CTORS = (1U << 0),
	BT_BF_DTORS = (1U << 1)
};

struct bt_process;

struct bt_bf_ctors_iterator {
	bt_bf_t *bf;
	struct bt_process *process;
	unsigned int ctors_index;
	unsigned int dtors_index;
	enum bt_bf_ctors_type flag;
	size_t ctors_size;
	size_t dtors_size;
	unsigned char *ctors_data;
	unsigned char *dtors_data;
};
typedef struct bt_bf_ctors_iterator bt_bf_ctors_iterator_t;

struct bt_bf_ctors_entry {
	enum bt_bf_ctors_type type;
	unsigned long long address;
};
typedef struct bt_bf_ctors_entry bt_bf_ctors_entry_t;

/*
 * Initializes an iterator for ctors/dtors section iterator.
 */
bool bt_bf_ctors_iterator_init(bt_bf_ctors_iterator_t *, bt_bf_t *,
    struct bt_process *, enum bt_bf_ctors_type);

/*
 * Deinitializes iterator object.
 */
void bt_bf_ctors_iterator_deinit(bt_bf_ctors_iterator_t *);

/*
 * Iterates over .ctors and .dtors sections providing the contents
 * to the user.
 */
enum bt_bf_iterator_result bt_bf_ctors_iterator_next(bt_bf_ctors_iterator_t *,
    bt_bf_ctors_entry_t *);

/*
 * Gets a binaries file header meta-data.
 */
void bt_bf_header_stat(bt_bf_t *, struct bt_bf_header_md *);

struct bt_bf_symbol_iterator {
	bt_bf_t *bf;
	unsigned int index;
};
typedef struct bt_bf_symbol_iterator bt_bf_symbol_iterator_t;

/*
 * Initializes iterator for the symbol cache
 */
void bt_bf_symbol_iterator_init(bt_bf_symbol_iterator_t *, bt_bf_t *);

/*
 * Iterator for the symbol cache
 */
enum bt_bf_iterator_result
bt_bf_symbol_iterator_next(bt_bf_symbol_iterator_t *, struct bt_bf_symbol *);

#endif /* _BT_BF_H */
