#ifndef _BT_BT_H
#define _BT_BT_H

#ifndef _GNU_SOURCE
#define _GNU_SOURCE 1
#endif

#ifndef ps_prochandle
#define ps_prochandle bt_process
#endif

#include <ck_queue.h>
#include <ck_hs.h>
#include <ck_ht.h>
#include <limits.h>
#include <pthread.h>
#include <unistd.h>
#include <signal.h>
#include <stdbool.h>
#include <stdint.h>
#include <thread_db.h>
#include <sys/queue.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/types.h>

#if defined(__sun__)
#include <proc_service.h>
#else
#include <sys/ptrace.h>
#include <sys/procfs.h>
#endif

#if defined(__FreeBSD__)
#include <proc_service.h>
#include <rtld_db.h>
#endif /* __FreeBSD__ */

#include "array.h"
#include "asset.h"
#include "bf.h"
#include "bt_unwind.h"
#include "dwarf.h"
#include "error.h"
#include "gvc.h"
#include "misc.h"
#include "os.h"
#include "register.h"
#include "so.h"
#include "target.h"
#include "tree.h"
#include "type.h"

/* Forward declaration for bt_pmap_entry. */
struct bt_pmap;
typedef struct bt_pmap bt_pmap_t;

#define BT_PMAP_MODE_NONE	0
#define BT_PMAP_MODE_EX		(1U << 0)
#define BT_PMAP_MODE_RD		(1U << 1)
#define BT_PMAP_MODE_WR		(1U << 2)
#define BT_PMAP_MODE_PR		(1U << 3)
#define BT_PMAP_MODE_SH		(1U << 4)
#define BT_PMAP_MODE_ANON	(1U << 5)
#define BT_PMAP_MODE_STACK	(1U << 6)
#define BT_PMAP_MODE_VVAR	(1U << 7)
#define BT_PMAP_MODE_VDSO	(1U << 8)
#define BT_PMAP_MODE_VSYSCALL	(1U << 9)
#define BT_PMAP_MODE_ALL	(UINT_MAX)

typedef unsigned int bt_pmap_mode_t;

struct bt_pmap_asset;
struct bt_pmap_entry {
	/* Base address. */
	unsigned long long base;

	/* Number of bytes from base address. */
	unsigned long long length;

	/* Base permissions. */
	bt_pmap_mode_t mode;

	/* Process map associated with entry. */
	bt_pmap_t *map;

	/* Asset associated with entry. */
	struct bt_pmap_asset *asset;

	/* Basename of path. Equivalent to path if not a file path. */
	const char *basename;

	unsigned long long offset;
};
typedef struct bt_pmap_entry bt_pmap_entry_t;

/* Forward declaration for bt_pmap. */
struct bt_process;
typedef struct bt_process bt_process_t;

struct bt_thread;
typedef struct bt_thread bt_thread_t;

struct bt_pmap_map {
	bt_bf_t *object;
	unsigned long long base;
};

#define BT_PMAP_F_LINKER_LOAD		(1UL << 0)
#define BT_PMAP_F_LINKER_LOAD_FAIL	(1UL << 1)

struct bt_pmap {
	/* Process associated with map. */
	bt_process_t *process;
	unsigned long flags;

	struct {
		/* Address to file cache. */
		bt_pmap_entry_t *address;
		unsigned int address_length;
		unsigned int address_capacity;

		/* Name to mapping cache. */
		ck_ht_t file;
	} cache;
};

/*
 * Regarding the map address ranges (such as from /proc/pid/maps)...
 * In some cases:
 * The maps entries have an end value equal to the last valid byte in
 * the mapping.  For instance, if a mapping started at 0x8000 and had
 * size 0x1000, then the entry would have an end value of 0x8fff.
 * In other observed cases:
 * The maps entries have an end value equal to the start of the next
 * possible mapping, i.e start-end: 0x1000 - 0x2000.
 * To handle both possible scenarios we will just add 1 and page mask
 * the end address.
 */
unsigned long long bt_pmap_align_end_addr(unsigned long long);

/*
 * Process map types:
 * - BT_PMAP_EXECUTABLE: Corresponds to executable regions.
 * - BT_PMAP_MEMORY: Corresponds to backing memory regions.
 */
enum bt_pmap_type {
	BT_PMAP_EXECUTABLE = 0,
	BT_PMAP_MEMORY,
};

#define BT_PMAP_FOREACH(C, P, M)					\
	for ((C) = NULL; bt_pmap_foreach_next(P, M, &(C)) == true;)

#define BT_LOG_ALL	(UINT_MAX)
#define BT_LOG_WARNING	(1U << 0)
#define BT_LOG_ERROR	(1U << 1)
#define BT_LOG_DEBUG	(1U << 2)

#define BT_CONFIG_INITIALIZER \
{ .log_mask = 0, .log = NULL, .asset_path = NULL, .load_assets = true, \
    .flags = 0, .map.ignore = NULL, .map.n_load = 0, .map.load = NULL }

/* Do not attempt to decompress compressed binary sections */
#define BT_CONFIG_NO_DECOMPRESS			(1UL << 0)
/* Do not use supplied debug address range data */
#define BT_CONFIG_NO_ARANGES			(1UL << 1)
/* Do not apply relocations in binary files */
#define BT_CONFIG_NO_RELOCATE			(1UL << 2)

struct bt_config_load {
	/*
	 * Name of object to load.  This can take one of the following values:
	 * -- NULL, which indicates to load the process executable
	 * -- An absolute path to an object file to load
	 * -- A full or partial object name.  For example, any the following
	 *    values would match an object with path /usr/lib/libjemalloc.so.1
	 *        libjemalloc
	 *        jemalloc
	 *        jemall
	 */
	const char *object_name;

	/*
	 * An array of CU names to load.  A NULL value is
	 * considered to represent an empty list.
	 */
	const char **cu_names;

	/*
	 * The length of the above array.
	 */
	unsigned int n_cu;
};

struct bt_config {
	unsigned int log_mask;
	void (*log)(unsigned int, const char *, ...);
	const char *asset_path;
	bool load_assets;
	unsigned long flags;

	struct {
		/*
		 * NULL-terminated list of paths to ignore when trying to read
		 * from memory.  A NULL pointer is considered to be an empty
		 * list.
		 */
		char **ignore;

		/*
		 * An array of structs specifying DWARF data to process
		 * at runtime.  See struct definition for more information.
		 * If this field is non-NULL, the n_load field must be set
		 * to the number of entries in the array.
		 */
		const struct bt_config_load *load;
		unsigned int n_load;
	} map;

	/*
	 * Optional: supply a hash seed to prevent libbt from opening
	 * /dev/urandom in order to get a hash seed. If set is true, seed is
	 * used, otherwise it is ignored.
	 */
	struct {
		bool set;
		uint64_t seed;
	} random;
};
extern struct bt_config bt_config;

bool bt_init(struct bt_config *);

#define bt_log(X, Z, ...) do {					\
	if ((X) & bt_config.log_mask) {				\
		bt_config.log(X,				\
		    ":[%s:%d] " Z,				\
		    __FUNCTION__, __LINE__,			\
		    ##__VA_ARGS__);				\
	}							\
} while (0)

size_t bt_md_pagesize(void);

struct bt_td {
	td_thragent_t *agent;
};
typedef struct bt_td bt_td_t;

struct bt_td_thread {
	td_thrhandle_t handle;
};
typedef struct bt_td_thread bt_td_thread_t;

struct bt_td_thread_st {
	td_thrinfo_t info;
	td_thrhandle_t handle;
};
typedef struct bt_td_thread_st bt_td_thread_st_t;

enum bt_thread_status {
	BT_THREAD_STATUS_UNKNOWN = 0,
	BT_THREAD_STATUS_RUNNING,
	BT_THREAD_STATUS_ZOMBIE,
	BT_THREAD_STATUS_IDLE,
	BT_THREAD_STATUS_SLEEPING,
	BT_THREAD_STATUS_WAITING,
	BT_THREAD_STATUS_STOPPED,
	BT_THREAD_STATUS_BLOCKED,
	BT_THREAD_STATUS_DEAD,
	BT_THREAD_STATUS_TRACED,
	BT_THREAD_STATUS_DISK,
	BT_THREAD_STATUS_QUEUED
};

/*
 * If this flag is set then unwinding will stop after this frame.  Otherwise
 * unwinding will continue as normal.
 */
#define BT_THREAD_UNWIND_F_STOP			(1U << 0)

/*
 * If this flag is set then the output frame will be appended to the existing
 * stack.
 */
#define BT_THREAD_UNWIND_F_APPEND		(1U << 1)

/*
 * If this flag is set then the output frame will replace the top frame in the
 * existing stack.  This flag is mutually exclusive with
 * BT_THREAD_UNWIND_F_APPEND.
 */
#define BT_THREAD_UNWIND_F_REPLACE		(1U << 2)

/*
 * Callback to be executed after each frame of a stack is unwound
 * The first argument represents the most recently unwound frame.
 * The second argument can be populated with a new register set to append to the
 * stack or replace the topmost frame with.
 * The fourth argument is a closure provided by the user to bt_thread.
 * The return value is a set of output flags indicating how unwinding should
 * proceed. See above for available flag values.
 */
typedef unsigned int bt_thread_unwind_cb_t(bt_unwind_cursor_t *,
    struct bt_register_set *, void *);

#define BT_THREAD_F_FAULT		(1UL << 0)

/*
 * A thread represents an execution context.
 */
struct bt_thread {
	bt_process_t *process;			/* Primary process. */
	pid_t pid;				/* Kernel process ID */
	pid_t tid;				/* Kernel thread ID */
	enum bt_thread_status status;
	struct {
		void *md;
	} target;
	bt_unwind_cursor_t cursor;
	bt_unwind_cursor_t *current;

	unsigned long flags;
	const char *name;

	bt_thread_unwind_cb_t *unwind_cb;
	void *unwind_closure;

	CK_LIST_ENTRY(bt_thread) process_linkage;
};

static inline bool
bt_thread_fault(const struct bt_thread *thread)
{

	return thread->flags & BT_THREAD_F_FAULT;
}

static inline const char *
bt_thread_name(const struct bt_thread *thread)
{

	return thread->name;
}

static inline enum bt_thread_status
bt_thread_status(const struct bt_thread *thread)
{

	return thread->status;
}

static inline void
bt_thread_unwind_cb_set(struct bt_thread *thread, bt_thread_unwind_cb_t *cb,
    void *closure)
{

	thread->unwind_cb = cb;
	thread->unwind_closure = closure;
	return;
}

/*
 * Determine the TLS base for the given module and write the result to the third
 * argument.
 *
 * The second argument is a pointer to the link map address for the current
 * module.  This may be modified during TLS computation.
 *
 * The output value is only valid if the return value is true.
 */
bool bt_thread_tls_base(bt_thread_t *, unsigned long long *,
    unsigned long long *, bt_error_t *);

enum bt_personality {
	BT_PERSONALITY_GDB = 0,
	BT_PERSONALITY_STRICT
};

struct bt_process_options {
	pid_t *tid;
	size_t n_tid;

	/*
	 * If this is non-NULL, libbt will attempt to load mapped object files
	 * from the provided directory (appending the absolute path specified
	 * by the target). If that fails, libbt will try again with the
	 * original absolute path.
	 *
	 * For instance, suppose that the file /foo/bar/baz.so is loaded in a
	 * target's process map and the user provides /some/path in the process
	 * options. The process will first attempt to read from
	 * /some/path/foo/bar/baz.so. If the fallback option is set then if
	 * that file cannot be found, it will then fall back to
	 * /foo/bar/baz.so.
	 */
	const char *map_path;

	/*
	 * If true, then if the map path fails to resolve a file, libbt will
	 * attempt with the original absolute path.
	 */
	bool map_path_fallback;

	struct bt_asset_config *asset;

	/*
	 * Size (in bytes) of the process read cache.
	 * A value of 0 will disable read caching.
	 */
	size_t cache_size;

	/*
	 * Debugger personality for target process.
	 */
	enum bt_personality personality;

	/*
	 * Unwind backend to use
	 */
	enum bt_unwind_type unwind_type;

	/*
	 * Maximum depth to search for tailcall chains.  A value of zero
	 * specifies unlimited depth.
	 */
	unsigned int tailcall_limit;
};

/*
 * TODO: These are currently a singleton as an interface refactor is required
 *       in order to propagate the value. Currently, bt_bf is entirely distinct
 *	 of process-level options, as it was originally designed for sharing
 *	 across processes.
 */
extern const char *bt_sysroot;
extern bool bt_sysroot_fallback;

#define BT_DEFAULT_TAILCALL_LIMIT 8

#define BT_PROCESS_OPTIONS_INITIALIZER {		\
	.tid = NULL,					\
	.n_tid = 0,					\
	.map_path = NULL,				\
	.map_path_fallback = false,			\
	.asset = NULL,					\
	.cache_size = 0,				\
	.personality = BT_PERSONALITY_GDB,		\
	.unwind_type = BT_UNWIND_BT,			\
	.tailcall_limit = BT_DEFAULT_TAILCALL_LIMIT	\
}

#define BT_PROCESS_INITIALIZER(x) \
	{ .pid = 0, .threads = CK_LIST_HEAD_INITIALIZER(x) }

struct bt_process_cache_block;

struct bt_process_cache_stat {
	/* Total number of requests received by the cache */
	unsigned long long requests;
	unsigned long long req_bytes;

	/* Number of requests satisfied by an existing cache entry */
	unsigned long long hits;
	unsigned long long hit_bytes;

	/* Number of requests satisfied by the most recently used cache entry */
	unsigned long long mru_hits;
	unsigned long long mru_bytes;

	/* Number of requests not satisfied by an existing cache entry */
	unsigned long long misses;
	unsigned long long miss_bytes;

	/* Number of requests unable to be satisfied by cache */
	unsigned long long errors;
	unsigned long long error_bytes;
};

struct bt_process_cache {
	/* Tree of populated cache entries sorted by memory address */
	RB_HEAD(bt_cache_tree, bt_process_cache_block) lookup;

	/*
	 * List of populated cache entries, ordered from least recently used to
	 * most recently used.
	 * This is represented as a doubly-linked tail queue to allow for
	 * the following operations in constant-time:
	 * -- Check for least recently used block
	 * -- Remove an arbitrary block from the list (given that we already
	 *    have a pointer to that block)
	 * -- Add a newly accessed block to the end of the list
	 */
	TAILQ_HEAD(, bt_process_cache_block) lru;

	/* List of unused cache entries */
	STAILQ_HEAD(, bt_process_cache_block) free;

	/*
	 * Cache the single most recently used address and data block
	 * to quickly satisfy bunched requests.
	 */
	struct {
		uintptr_t address;
		unsigned char *data;
	} mru;

	unsigned char *data;
	struct bt_process_cache_block *blocks;
	size_t block_size;
	unsigned int n_block;

	struct bt_process_cache_stat stat;
};
typedef struct bt_process_cache bt_process_cache_t;

#define BT_PROCESS_R_MAP	(1UL << 0)

struct bt_process_cu_subprograms;

struct bt_process {
	unsigned long flags;

	/* Primary PID. */
	pid_t pid;

	struct bt_process_options options;

	/* Resources initialized. */
	unsigned long resources;

	/* Primary executable file */
	bt_bf_t *executable;

	/* Path of executable file */
	const char *exec_path;

	/*
	 * A map of executable assets associated with the process.
	 * This allows for both address to object look-up as well
	 * as path to object look-up. A map is also able to specify
	 * a path prefix.
	 */
	bt_pmap_t map;

	struct {
		bt_target_t *target;
		void *md;
	} target;

	/* Unwind state. */
	bt_unwind_t unwind;

	bt_unwind_cursor_t cursor;

	/*
	 * List of threads. We use non-blocking single-writer variant to allow
	 * for signal-safe management of the list.
	 *
	 * The thread map itself is a mapping from PID to thread object.
	 */
	ck_hs_t thread_map;
	CK_LIST_HEAD(, bt_thread) threads;
	unsigned int n_threads;

	/*
	 * If a primary faulting thread is found, it is cached here.
	 */
	struct bt_thread *faulted;

	bt_process_cache_t read_cache;

	/* DWARF call site cache */
	struct {
		ck_hs_t cache;
	} dwarf_call_site;

	/* Environment variable cache */
	struct {
		ck_hs_t cache;
	} environ;

	struct {
		/*
		 * Processed CU cache; entries have already been parsed and
		 * cached by the process object (e.g. for call sites and types).
		 */
		ck_hs_t cache;

		/* Type name cache. */
		ck_hs_t typenames;

		/*
		 * Subprograms sorted by pc range.
		 * CU subprograms are separated (i.e. through list entries) to
		 * avoid the cost of repeatedly sorting a large subprogram
		 * array (we will have terrible sort performance if we add a
		 * new CU with DIE ranges < existing ranges).
		 * CUs may not have low-high PC ranges, so we use a list here
		 * rather than the typical RB tree.
		 */
		STAILQ_HEAD(, bt_process_cu_subprograms) subprograms;
	} cu;
};

static inline enum bt_target_type
bt_process_target_type(const struct bt_process *process)
{

	return bt_target_type(process->target.target);
}

void bt_process_cache_stat(struct bt_process_cache_stat *,
    const bt_process_t *);

/*
 * Attach to a target and populate the process with data from that target.
 */
bool bt_process_attach(bt_process_t *, bt_target_t *,
    struct bt_process_options *, bt_error_t *);

/*
 * Parse the specified CU for cachable data (e.g. callsites and types) and
 * store it in the specified process.
 */
void bt_process_cu_cache(bt_process_t *, struct bt_dwarf_cu *);

/*
 * Caches the type specified by the given DIE in the given process. The bt_type
 * is optional and may be NULL if not already known.
 */
bool bt_process_cache_type(bt_process_t *, struct bt_dwarf_die *,
    struct bt_type *);

/*
 * Caches the subprogram specified by the given DIE in the the given process.
 */
bool bt_process_cache_subprogram(struct bt_process_cu_subprograms *,
    struct bt_dwarf_die *);

/*
 * Retrieve the type associated with the specified type name. Only CUs the
 * specified process has already encountered and parsed are considered. An
 * optional CU filter may be specified; if so, only types belonging to the
 * specified CU will be considered. Returns NULL on failure.
 */
struct bt_type *bt_process_typename_get(bt_process_t *,
    const char *, struct bt_dwarf_cu *, bt_error_t *);

/*
 * Retrieves all types known.
 * It returns an array of struct bt_type*, or NULL on failure.
 * Caller is responsible to release the array.
 */
struct bt_type **bt_process_types(bt_process_t *, size_t*, bt_error_t *);

/*
 * Retrieve the cached subprogram DIE associated with the given pc. If CU is
 * specified, only that particular CU will be checked (if already cached).
 * Returns NULL on failure.
 */
struct bt_dwarf_die *bt_process_subprogram_get(bt_process_t *,
    struct bt_dwarf_cu *, unsigned long long);

/*
 * Detach from a target.  This will not destroy any snapshot data and the
 * snapshot will be usable until it is deinitialized.
 */
bool bt_process_detach(bt_process_t *, unsigned long, bt_error_t *);

/*
 * Ensure that the process with the provided ID is in a stopped state but stop
 * trying after the specified timeout interval.
 * A timeout interval of 0 indicates that a blocking wait should be performed.
 * Returns true if the process was stopped and false otherwise.
 */
bool bt_process_stop(bt_process_t *, pid_t, unsigned long, bt_error_t *);

/*
 * Destroy all data in a snapshot.
 */
void bt_process_deinit(bt_process_t *);

/*
 * If the underlying target has identified a primary faulting thread, then
 * this function returns true. Software can use this thread as the basis
 * for grouping and classification. Otherwise, consumers of libbt may wish
 * to treat threads with asynchronous events as they please.
 */
bt_thread_t *bt_process_fault_get(const bt_process_t *);

/*
 * Set the faulting thread for the process.
 */
void bt_process_fault_set(bt_process_t *, bt_thread_t *);

static inline const char *
bt_process_exec_path(const struct bt_process *process)
{

	return process->exec_path;
}

static inline enum bt_personality
bt_process_personality(const struct bt_process *process)
{

	return process->options.personality;
}

/*
 * Read bytes from a process' memory into the provided buffer.
 *
 * Returns the number of bytes read on success and -1 otherwise.
 */
ssize_t bt_process_read(bt_process_t *, unsigned long long, void *, size_t,
    bt_error_t *);

/*
 * Stop the target process.
 */
bool bt_process_pstop(bt_process_t *, unsigned long, bt_error_t *);

/*
 * Continue the target process.
 */
bool bt_process_pcontinue(bt_process_t *, bt_error_t *);

/*
 * Synthesizes a thread from its initial register set.
 */
bool bt_process_thread_synthesize(bt_process_t *, pid_t, pid_t,
    const char *name, struct bt_register_set *, bt_thread_unwind_cb_t *,
    void *, bt_error_t *);

/*
 * Associate a pre-instantiated thread object with a process.
 */
bool bt_process_thread_add(bt_process_t *, bt_thread_t *);

/*
 * Removes a thread from the process it is associated with.
 */
void bt_process_thread_remove(bt_thread_t *);

/*
 * Attaches to a thread with the given thread ID.
 */
bool bt_process_thread_attach(bt_process_t *, pid_t, bt_error_t *);

/*
 * Detaches from a thread with the given thread ID.
 */
bool bt_process_thread_detach(bt_thread_t *, unsigned long, bt_error_t *);

/*
 * Stop the thread with the given ID.
 */
bool bt_process_thread_stop(bt_process_t *, pid_t, bt_error_t *);

/*
 * Continue the thread with the given ID.
 */
bool bt_process_thread_continue(bt_process_t *, pid_t, bt_error_t *);

/*
 * Read information for the thread with the given thread ID into the provided
 * buffer.
 */
bool bt_process_thread_info(bt_process_t *, pid_t, void *, bt_error_t *);

/*
 * Write the CPU index for the provided thread into the provided buffer.
 * If the return value is false then the contents of the buffer are invalid.
 */
bool bt_process_thread_cpu(bt_thread_t *, unsigned int *, bt_error_t *);

enum bt_process_siginfo_result {
	BT_PROCESS_SIGINFO_SUCCESS,	/* Signal information found */
	BT_PROCESS_SIGINFO_FAILURE,	/* Error occurred reading signal information */
	BT_PROCESS_SIGINFO_NONE		/* No signal information present */
};

/*
 * Read signal information from a thread.
 */
enum bt_process_siginfo_result bt_process_thread_siginfo(bt_thread_t *,
    siginfo_t *, bt_error_t *);

/*
 * Populate the output arguments with information about the debug file for use
 * with the address specified in the second argument.
 *
 * This is primarily for use with libunwind.
 */
bool bt_process_get_elf_image(bt_process_t *, unsigned long long,
    unsigned long *, unsigned long *, char *, size_t, bt_error_t *);

/*
 * Find the binary file with debug information for the provided address.
 * Returns NULL if no such file exists.
 */
bt_bf_t *bt_process_backing_file(bt_process_t *, unsigned long long);

/*
 * Find the binary file with debug information for the path provided in the
 * second argument.  The third argument will be populated with the original name
 * of that file.
 *
 * The string stored in the third argument is managed by the target and the
 * caller should not modify or free it.
 * Returns NULL if no such file exists.
 */
bt_bf_t *bt_process_backing_file_path(bt_process_t *, const char *,
    const char **);

/*
 * Returns a function for performing specialized unwinding of a frame at the
 * given address.  If the frame should be unwound normally, this function
 * returns NULL.
 */
bt_unwind_fn_t *bt_process_get_unwinder(bt_process_t *, unsigned long long,
    void **);

/*
 * Unwind a stack frame using the provided cursor.
 */
enum bt_unwind_result bt_process_unwind(bt_unwind_cursor_t *, bt_error_t *);

/*
 * Populate a thread with the provided thread_db data (passed as a void
 * pointer).
 */
bool bt_process_td_thread_populate(bt_process_t *, pid_t,
     bt_td_thread_t *, bt_error_t *);

/*
 * Attach to threads that aren't already included in the process
 */
#define BT_TD_F_ATTACH			(1U << 0)

/*
 * Initialize thread debugging interface.
 */
bool bt_td_init(bt_td_t *, bt_process_t *, unsigned int, bt_error_t *);

/*
 * Populate the fourth argument with the TLS base of the given module and
 * thread.
 *
 * The third argument is a pointer to the module link map.  It may be modified.
 *
 * The TLS base value is only valid if the return value is true.
 */
bool bt_td_tls_base(bt_thread_t *, bt_td_thread_t *, unsigned long long *,
    unsigned long long *, bt_error_t *);

/*
 * Initialize a thread using an opaque handle provided by thread_db
 */
bool bt_td_thread_init(const void *, bt_thread_t *, bt_error_t *);

/*
 * Deinitialize thread debugging interface.
 */
void bt_td_deinit(bt_td_t *);

#define BT_PROCESS_THREAD_FIRST(P)				\
	CK_LIST_FIRST(&(P)->threads)

#define BT_PROCESS_FOREACH_THREAD(P, T)				\
	CK_LIST_FOREACH(T, &(P)->threads, process_linkage)

#define BT_PROCESS_FOREACH_THREAD_SAFE(P, T, N)				\
	for ((T) = CK_LIST_FIRST(&(P)->threads);			\
	     (T) && ((N) = CK_LIST_NEXT((T), process_linkage),1);	\
	     (T) = (N))

struct bt_thread *bt_process_thread_get(struct bt_process *, pid_t);

static inline unsigned int
bt_process_thread_count(struct bt_process *process)
{

	return process->n_threads;
}

static inline pid_t
bt_process_pid(bt_process_t *process)
{

	return process->pid;
}

struct bt_asset;

/*
 * Returns 1 if specified pid is targeted
 * on attach. It will only return 1 for every target.
 */
int bt_process_include(bt_process_t *, pid_t);

/* Detach without memory clean-up. */
#define BT_PROCESS_DETACH_FAST		(1U << 0)

/* Detach from process with SIGSTOP pending. */
#define BT_PROCESS_DETACH_STOP		(1U << 1)

/* Detach from process with SIGCONT pending. */
#define BT_PROCESS_DETACH_CONT		(1U << 2)

struct bt_process_environ_entry;

/*
 * Retrieves the environment variables from a given target.
 * It returns an array of struct bt_process_environ_entry
 * containing key/value for each variable.
 */
struct bt_process_environ_entry *bt_process_environ(bt_process_t *,
    size_t *count, bt_error_t *);

/*
 * Initializes a pmap.
 */
bool bt_pmap_init(bt_pmap_t *, bt_process_t *);

/*
 * Fetches a pmap for a process.
 * Returns NULL if no such pmap is available for the process.
 * May never return NULL if type is BT_PMAP_EXECUTABLE or BT_PMAP_MEMORY.
 */
bt_pmap_t *bt_pmap_get(bt_process_t *, enum bt_pmap_type);

/*
 * Destroys a pmap object.
 */
void bt_pmap_deinit(bt_pmap_t *);

bool bt_pmap_map(bt_pmap_t *,
    bt_pmap_mode_t,
    const char *,
    unsigned long long base,
    unsigned long long length,
    unsigned long long offset);
bool bt_pmap_const(bt_pmap_t *);

/*
 * Implements address to map entry look-up.
 */
bt_pmap_entry_t *bt_pmap_find(bt_pmap_t *, unsigned long long);

/*
 * Implements name to symbol look-up.
 */
bt_bf_symbol_t *bt_pmap_symbol_name(bt_pmap_t *, const char *, const char *,
    const char **, bt_bf_t **, unsigned long long *);

/*
 * Implements name to file look-up.
 */
bt_bf_t *bt_pmap_bf_path(bt_pmap_t *, const char *, const char **);

#define BT_PMAP_BF_PARTIAL		(1UL << 0)

/*
 * Find a binary file based on object name.
 * If the BT_PMAP_BF_PARTIAL flag is supplied, this supports partial matches,
 * so, for example, a user could find the file /usr/lib/.../libjemalloc.so.1
 * with the queries "libjemalloc", "jemalloc", "jem", etc.
 * Otherwise, only exact basename matches are supported.
 * If the last argument is not NULL, it will be populated with the original
 * (non-aliased) name of the file.
 */
bt_bf_t *bt_pmap_bf_name(bt_pmap_t *, const char *, unsigned long,
    const char **);

/*
 * Find a binary file based on process address.
 */
bt_bf_t *bt_pmap_bf_addr(bt_pmap_t *, unsigned long long);

/*
 * Indicate that any pmap entries with the path specified in the second
 * argument should instead use the backing file specified by the path in the
 * third argument.
 */
bool bt_pmap_alias_set(bt_pmap_t *, const char *, const char *);

/*
 * Returns the path for a pmap entry.
 */
const char *bt_pmap_entry_path(const bt_pmap_entry_t *);

/*
 * Get the next address in the cache for the requested process map.
 * Used primarily within BT_PMAP_FOREACH.
 */
static inline bool
bt_pmap_foreach_next(bt_process_t *process, enum bt_pmap_type pt,
    bt_pmap_entry_t **entry)
{
	bt_pmap_t *pmap = bt_pmap_get(process, pt);

	if (pmap == NULL)
		return false;

	if (*entry == NULL)
		*entry = &pmap->cache.address[0];
	else
		(*entry)++;

	return *entry < &pmap->cache.address[pmap->cache.address_length];
}

/*
 * Returns pointer to process associated with the specified handle.
 */
bt_process_t *bt_process_unw(void *);

/*
 * Returns pointer to error string.
 */
const char *bt_strerror(int);

#define BT_RESOLVE_PMAP			(1UL << 0)
#define BT_RESOLVE_SYMBOL		(1UL << 1)
#define BT_RESOLVE_LINE			(1UL << 2)
#define BT_RESOLVE_FRAME		(1UL << 3)
#define BT_RESOLVE_VARIABLE		(1UL << 4)
#define BT_RESOLVE_INLINE		(1UL << 5)
#define BT_RESOLVE_ARANGES		(1UL << 6)
#define BT_RESOLVE_ALL			(ULONG_MAX)

#define BT_FRAME_F_RESOLVED		(1UL << 0)
#define BT_FRAME_F_NO_RESOLVE		(1UL << 1)
#define BT_FRAME_F_TAILCALL_CALLER	(1UL << 2)
#define BT_FRAME_F_TAILCALL_CALLEE	(1UL << 3)
#define BT_FRAME_F_TAILCALL_UNINCLUDED	(1UL << 4)
#define BT_FRAME_F_UNWIND_HEURISTIC	(1UL << 5)

#define BT_FRAME_F_GUESSED \
    (BT_FRAME_F_TAILCALL_CALLER | BT_FRAME_F_UNWIND_HEURISTIC)

struct bt_frame {
	const char *name;
	unsigned long line;

	unsigned long flags;

	/*
	 * Inline entity mapping. Presence of these values indicate
	 * that the location is an instance of an inlined subroutine.
	 */
	struct {
		unsigned long long file;
		unsigned long long line;
	} il;

	unsigned long long cfa;
	unsigned long long frame_base;
	unsigned int fprm;
	struct bt_unwind_cursor cursor;
	union {
		SLIST_HEAD(, bt_frame) head;
		SLIST_ENTRY(bt_frame) linkage;
	} inlined;
	STAILQ_HEAD(, bt_variable) variables;
	struct bt_frame *next;
	struct bt_frame *previous;

	/* Frame error context. */
	bt_error_t error;

	/*
	 * For inlined frames, this points to the containing frame.
	 */
	struct bt_frame *parent;
};
typedef struct bt_frame bt_frame_t;

static inline const char *
bt_frame_error(const struct bt_frame *f)
{

	return bt_error_message(&f->error);
}

static inline const char *
bt_frame_name(const struct bt_frame *f)
{

	return f->name;
}

/*
 * Later, will return declaration information.
 */
static inline unsigned long
bt_frame_decl(const struct bt_frame *f)
{

	return f->line;
}

static inline bool
bt_frame_has(const struct bt_frame *frame, unsigned int flags)
{

	return frame->flags & flags;
}

static inline bool
bt_frame_is_signal(struct bt_frame *frame)
{

	return bt_unwind_cursor_has(&frame->cursor, BT_UNWIND_SIGNAL);
}

static inline bool
bt_frame_contains_signal(struct bt_frame *frame)
{

	return bt_unwind_cursor_has(&frame->cursor, BT_UNWIND_SIGNAL_CONTAINED);
}

/*
 * Return the current non-inlined stack frame.
 */
static inline bt_frame_t *
bt_frame_base(bt_frame_t *f)
{
	bt_frame_t *parent = f;

	while (parent->parent != NULL)
		parent = parent->parent;

	return parent;
}

/*
 * Attempts to recover siginfo structure from frame.
 */
enum bt_process_siginfo_result bt_frame_siginfo(siginfo_t *, bt_frame_t *);

/*
 * Returns the parameter at the specified argument offset,
 * starting from 0.
 */
bt_variable_t *bt_frame_fprm(bt_frame_t *, unsigned int);

struct bt {
	struct bt_frame *frame;
	size_t n_frame;
};
typedef struct bt bt_t;

#define BT_FOREACH_FRAME(F, B)	\
	for ((F) = (B)->frame; (F); (F) = (F)->next)

#define BT_FOREACH_INLINE(F, P)	\
	SLIST_FOREACH(F, &(P)->inlined.head, inlined.linkage)

/*
 * Third argument is maximum number of frames.
 */
bool bt_thread(bt_t *, bt_thread_t *, size_t);

/*
 * Deinitialize backtrace object.
 */
void bt_deinit(bt_t *);

struct bt_resolver {
	struct bt_frame *frame;
	unsigned long resolved;
	unsigned long address;
	bt_bf_symbol_t *symbol;
	bt_pmap_entry_t *pe;
	struct bt_dwarf_line line;
};
typedef struct bt_resolver bt_resolver_t;

bool bt_resolver_line(bt_resolver_t *, bt_frame_t *,
    const char **, const char **,
    unsigned long long *);

#define BT_FRAME_FOREACH_VARIABLE(C, H) \
	STAILQ_FOREACH(C, &((H)->variables), linkage)

/*
 * Resolves frame to relevant data.
 */
unsigned long bt_resolve_frame(bt_resolver_t *, bt_frame_t *,
    unsigned long);

#define BT_GV_F_GVC		(1UL << 0)
#define BT_GV_F_SYMTAB		(1UL << 1)

/*
 * Not all compilers emit proper debug information for TLS variables.
 * If this flag is set, then the user has manually specified that a variable
 * is thread-local, even if the debug information does not reflect that.
 */
#define BT_GV_F_TLS		(1UL << 2)

/*
 * If no matches are found in on-disk global variable caches, attempt to load
 * global variable data at runtime from objects that were specified during
 * configuration.  This flag requires that BT_GV_F_GVC is also provided.
 */
#define BT_GV_F_GVC_RUNTIME	(1UL << 3)

/*
 * Do not fail if the object cannot be resolved.
 */
#define BT_GV_F_OPTIONAL	(1UL << 4)

struct bt_gv_query {
	const char *symbol;
	const char *object;
	const char *cu;
	void *closure;
	bt_variable_t **populate;
	unsigned long long flags;
};

#define BT_GV_QUERY_INITIALIZER \
	{ .symbol = NULL, .object = NULL, .cu = NULL, .closure = NULL,	\
	.flags = 0 }

enum bt_gv_result {
	BT_GV_SUCCESS,
	BT_GV_FAILURE
};

struct bt_gv_resolver_entry {
	bt_variable_t *var;
	void *closure;
};
typedef struct bt_gv_resolver_entry bt_gv_resolver_entry_t;

BT_ARRAY(gv_resolver);

struct bt_gv_resolver {
	BT_ARRAY(gv_resolver) *global;
	BT_ARRAY(gv_resolver) *tls;
	unsigned int count;
};
typedef struct bt_gv_resolver bt_gv_resolver_t;

struct bt_gv_resolver_iterator {
	bt_gv_resolver_t *resolver;
	BT_ARRAY(gv_resolver) *items;
	unsigned long flags;
	unsigned long state;
	unsigned int ix;
};
typedef struct bt_gv_resolver_iterator bt_gv_resolver_iterator_t;

#define BT_GV_RESOLVE_BODY(A, P, F, C, ...) ({			\
    const struct bt_gv_query q_##C[] = { __VA_ARGS__ };		\
    bt_gv_resolve(A, P, q_##C, BT_STATIC_ARRAY_SIZE(q_##C), F); \
})

#define BT_GV_RESOLVE_STUB(A, P, F, C, ...) \
	BT_GV_RESOLVE_BODY(A, P, F, C, __VA_ARGS__)

/*
 * Convenience macro allowing the passing of raw query structs.
 * The first argument is a pointer to a bt_gv_resolver struct to be populated.
 * The second argument is a pointer to the process being evaluated.
 * The third argument is the set of flags to use for resolution.  See
 *     bt_gv_resolve documentation for more information.
 * These arguments are followed by any number of bt_gv_query structs to resolve.
 */
#define BT_GV_RESOLVE(A, P, F, ...) \
	BT_GV_RESOLVE_BODY(A, P, F, __COUNTER__, __VA_ARGS__)

/*
 * Populate the provided resolver struct with global and thread-local variables
 * matching the provided query parameters.  The fourth argument indicates the
 * size of the query array and the fifth argument is used to provide control
 * flags.
 *
 * Each query may specify the following criteria (all of which are optional):
 *	symbol: If non-NULL, the query will only match variables whose names
 *		match the provided string.  If NULL, all variables will match
 *		this filter.
 *	object: If non-NULL, the query will only match variables contained in
 *		objects whose names match the provided string.  If NULL, all
 *		variables will match this filter.
 *	cu:     If non-NULL, the query will only match variables contained in
 *		compilation units whose names match the provided string.  If
 *		NULL, all variables will match this filter.
 *	closure: This pointer will be included with all iterator results
 *		matching this query.  It is not used by the resolver itself but
 *		is useful in cases where users want to associate some type of
 *		context with different queries.
 *	populate: If this pointer is non-NULL, then its location will be
 *		populated with the value of the last variable matching the
 *		query (or NULL if no variables match the query).
 *	flags:  This is a mask of control flags to determine different aspects
 *		of resolver behavior.  See all flags beginning with BT_GV_F_
 *		for more information about specific flags.
 */
enum bt_gv_result bt_gv_resolve(bt_gv_resolver_t *, bt_process_t *,
    const struct bt_gv_query *, size_t, unsigned long long);

/*
 * Returns the number of results from a resolver object that was initialized
 * with a call to bt_gv_resolve and returned success.
 */
unsigned int bt_gv_resolve_count(const bt_gv_resolver_t *);

/*
 * The resolver interface above will return a single instance of each TLS
 * variable which is not associated with any thread.  The caller is responsible
 * for calling this function on each thread they are interested in in order to
 * get TLS variable data for those threads.
 * The first argument is the variable returned by bt_gv_resolve, the second
 * argument is a pointer to a variable struct to be populated (whose lifetime
 * the caller is responsible for managing), and the third argument is the thread
 * to use for evaluating the variable.
 */
bool bt_gv_resolve_variable_tls(bt_variable_t *, bt_variable_t *,
    bt_thread_t *, bt_error_t *);

void bt_gv_resolver_deinit(bt_gv_resolver_t *);

bool bt_gv_register(const struct bt_config_load *);

#define BT_GV_ITERATOR_F_GLOBAL		(1UL << 0)
#define BT_GV_ITERATOR_F_TLS		(1UL << 1)

/*
 * Initialize a global variable iterator from a global variable resolver.
 * This iterator will only yield variables specified by the flags passed to the
 * second argument (global, TLS or both).
 */
void bt_gv_resolver_iterator_init(bt_gv_resolver_iterator_t *,
    bt_gv_resolver_t *, unsigned long);

/*
 * Fetch the next variable from a global variable iterator.
 * A return value of false indicates that all variables have already been
 * yielded.
 * Upon a return value of true, the second argument will be populated with a
 * pointer to a global variable and if the third argument is not NULL then it
 * will be populated with the pointer that the user attached to the query which
 * yielded this variable.
 */
bool bt_gv_resolver_iterator_next(bt_gv_resolver_iterator_t *,
    bt_variable_t **, void **);

/*
 * Finds all variables that have full type information and adds them to
 * the process' or threads' global variable lists.
 * The caller can specify which variables they are interested in via the string
 * arguments:
 *	The second argument is the name of the object file to search.  If this
 *		argument is NULL then all object files will be searched.
 *	The third argument is the name of the compilation unit to search.
 *		If this argument is NULL then all compilation units will be
 *		searched.
 *	The fourth argument is the name of the symbol to search for.  If this
 *		argument is NULL then all symbols from the specified object
 *		file and compilation unit will be included.
 */
bool bt_resolve_globals(bt_process_t *, const char *, const char *,
    const char *, unsigned long);

/*
 * Address to symbol name look-up.
 */
const char *bt_symbol_name_address(bt_unwind_cursor_t *, unsigned long long);
bt_bf_t *bt_pmap_pe_bf(bt_pmap_entry_t *);
const char *bt_pmap_pe_path(bt_pmap_entry_t *);

/*
 * Examine linker map metadata to find link map addresses for all dynamic
 * libraries.  This is required in order to provide TLS support for dynamic
 * libraries.
 */
bool bt_pmap_linker_load(bt_pmap_t *);

void bt_frame_init(bt_frame_t *, bt_unwind_cursor_t *);
bt_variable_t *bt_frame_variable_find(bt_frame_t *, const char *);
bt_variable_t *bt_process_variable_find(bt_process_t *, const char *);
bt_variable_t *bt_thread_variable_find(bt_thread_t *, const char *);

/*
 * This function reads a binary section from memory or a corefile.
 */
void *bt_process_section_read(bt_process_t *, bt_bf_t *,
    const char *, size_t *);

/*
 * accessor for telling whether or not a given pmap entry is an anonymous
 * type memory mapping (versus file mapping).
 */
bool bt_pmap_entry_anon(bt_pmap_entry_t *);

/*
 * flags for instructing the pmap iterator
 */
enum bt_pmap_iter_type {
	BT_PMAP_FILE = 		(1U << 0),
	BT_PMAP_ANON = 		(1U << 1),
	BT_PMAP_UNIQUE = 	(1U << 2),
	BT_PMAP_READ_CHUNK = 	(1U << 3),
};

enum bt_pmap_iterator_result {
	BT_PMAP_ITER_OK,
	BT_PMAP_ITER_DONE,
	BT_PMAP_ITER_ERROR
};

/*
 * The iterator type for iterating over pmap entries.
 * NOTE: We are using ht_iterator to be backwards compatible with the ck_ht
 * based cache we are accessing in the pmaps interface.
 */
struct bt_pmap_iterator {
	struct bt_process *process;
	struct bt_pmap *pmap;
	enum bt_pmap_iter_type type;
	ck_ht_iterator_t ht_iterator;
	ck_ht_t *cachefile;
	ck_ht_entry_t *ht_entry;
	size_t chunk_size;
	unsigned int index;
};
typedef struct bt_pmap_iterator bt_pmap_iterator_t;

struct bt_pmap_iterator_entry {
	bt_bf_t *bf; 		/* for MAP_FILE binary/executable segments */
	uint8_t *chunk; 	/* for  MAP_ANONYMOUS segments		   */
	size_t chunk_size;	/* initially set to iter->chunk_size	   */
	char *path; 		/* for MAP_FILE binary/non-binary segments */
	unsigned long long base;/* base address				   */
	size_t map_size; 	/* mapping size				   */
	bt_pmap_mode_t mode;	/* permissions, mapping type, file or anon */
};

/*
 * Iterate over file mappings and anonymous mappings.
 * Functionality is specified by the flags used. If only a single type
 * is set in the flag, then the iterator will only return entries
 * that are exclusive to the requested type.
 * If the caller requests file mappings and anonymous mappings, they
 * can discern whether it is anonymous by checking the pmap mode
 * flag for the presence of BT_PMAP_MODE_ANON.
 *
 * - BT_PMAP_FILE
 *	Returns file mappings...
 *	The iterator will return the base address, the segment size,
 * 	the mode, the path to the mapped file, and a corresponding bf if the
 *	file is of a supported binary format. Note that the entries mode field
 *	will be set to 0 if the BT_PMAP_UNIQUE flag is set and the entry yielded
 *	is not an anonymous mapping.
 *
 *- BT_PMAP_ANON
 *	Returns anonymous memory mappings... and if used with BT_PMAP_CHUNK_READ
 *	The iterator will set entry.chunk to a chunk of memory containing
 * 	data from the memory mapping that is chunk_size in bytes.
 *	If the chunk_size requested is larger than the memory
 *	memory mapping, then entry.chunk_size will be set to the new size.
 *
 *	-- example code --
 *	bt_pmap_iterator_entry_init(&entry);
 * 	bt_pmap_iterator_init(&iter, process, 128, PT_PMAP_ANON);
 *	entry.chunk = malloc(128); // optionally allocate the memory yourself
 *	bt_pmap_iterator_next(&iter, &entry);
 *	uint8_t *buf = alloca(entry.chunk_size);
 *	memcpy(buf, entry.chunk, entry.chunk_size);
 *
 *	-- note --
 *	If we hadn't allocated entry.chunk manually before calling iterator_next
 *	then the iterator would have allocated it to 128 bytes for us, in which
 *	case we would have been responsible for calling free(entry.chunk) after
 *	each successful iteration. There is no deinit function for pmap iterator
 *	since the only thing that would ever need freeing is chunk, and that varies
 *	based on whether or not the user allocated the memory for it or not.
 *
 *- BT_PMAP_UNIQUE
 *	If used in conjunction with BT_PMAP_FILE, then it will yield only
 *	a single unique entry for each file mapping. This can be useful if
 *	for instance, you want to get the path of every shared library that
 * 	is mapped into the process, without seeing duplicate path names since
 *	each shared library is likely to be present in several memory mappings
 *	for each of the libaries segments. Using this flag does void some aspects
 *	of the entry values that are yielded. With this flag you will only see
 *	the following entry values filled in: base, bf, and path.
 *
 * When initializing the pmap iterator with the BT_PMAP_ANON flag make sure to
 * specify a chunk_size (3rd argument to init function).
 */
bool bt_pmap_iterator_init(bt_pmap_iterator_t *, bt_process_t *,
    size_t, enum bt_pmap_iter_type);

/*
 * Initialize the pmap iterator entry
 */
void bt_pmap_iterator_entry_init(struct bt_pmap_iterator_entry *);

/*
 * Used to iterate over the pmaps. Should only be called after
 * bt_pmap_iterator_init, and bt_pmap_iterator_entry_init have been called.
 */
enum bt_pmap_iterator_result bt_pmap_iterator_next(bt_pmap_iterator_t *,
    struct bt_pmap_iterator_entry *, bt_error_t *);

/*
 * accessor to return the executable bf for a given process.
 */
static inline bt_bf_t *
bt_process_executable(struct bt_process *process)
{

	return process->executable;
}

/*
 * Iterator struct for getting the environment variables.
 */
struct bt_process_environ_iterator {
	size_t vcount;
	unsigned int index;
	struct ck_hs_iterator hs_iter;
	struct bt_process *process;
	struct bt_process_environ_entry *environ;
};
typedef struct bt_process_environ_iterator bt_process_environ_iterator_t;

struct bt_process_environ_entry {
	char *key; /* variable name */
	char *value; /* variable value */
};
typedef struct bt_process_environ_entry bt_process_environ_entry_t;

static inline char *
bt_process_environ_entry_key(struct bt_process_environ_entry *e)
{

	return e->key;
}

static inline char *
bt_process_environ_entry_value(struct bt_process_environ_entry *e)
{

	return e->value;
}

enum bt_environ_iterator_result {
	BT_ENVIRON_ITER_OK,
	BT_ENVIRON_ITER_ERROR,
	BT_ENVIRON_ITER_DONE
};
/*
 * iterate through each environment variable name=value pair
 */
enum bt_environ_iterator_result
bt_process_environ_iterator_next(struct bt_process_environ_iterator *,
    struct bt_process_environ_entry *, bt_error_t *);

/*
 * Make sure to call this after the bt_process_environ_iterator_init and
 * bt_process_environ_iterator_next functions have been used to free up
 * any associated memory.
 */
void bt_process_environ_iterator_deinit(struct bt_process_environ_iterator *);

/*
 * Initialize the iteration state for accessing environment variables
 */
bool bt_process_environ_iterator_init(struct bt_process_environ_iterator *,
    bt_process_t *);

/*
 * Lookup environment variable values by the variable name.
 */
bool bt_process_environ_lookup(bt_process_t *, const char *,
    char **, bt_error_t *);

/*
 * Looks up symbols from the entire process, and not just a single
 * binary.
 */
struct bt_bf_symbol *bt_process_symbol_lookup(bt_process_t *, const char *);

/*
 * Returns true if the entry has read permissions
 */
bool bt_pmap_entry_readable(bt_pmap_entry_t *);

/*
 * Returns true if the entry is of an anonymous mapping
 */
bool bt_pmap_entry_anon(bt_pmap_entry_t *);

/*
 * Resolves line and symbol information associated with the address.
 */
bool bt_process_resolve_pc(bt_process_t *, unsigned long,
    struct bt_dwarf_line *, const char **, unsigned long, unsigned long);

enum bt_md_utsname {
	BT_MD_SYSNAME,
	BT_MD_NODENAME,
	BT_MD_RELEASE,
	BT_MD_VERSION,
	BT_MD_MACHINE
};

const char * bt_md_uname(enum bt_md_utsname);

/*
 * Returns true if a pmap entry with BT_PMAP_STACK_MODE was found
 * and sets the low, and high value of the stack address range.
 */
bool
bt_pmap_stack_boundaries(bt_process_t *, unsigned long long *,
	unsigned long long *);
#endif /* _BT_BT_H */
