#ifndef _BT_GVC_H
#define _BT_GVC_H

#include <stdbool.h>
#include <sys/types.h>

#include "ac.h"
#include "array.h"

#define BT_GVC_VERSION	(1UL)

struct bt_gvc_section {
	unsigned char *start;
	unsigned int n_entry;
};

struct bt_asset;
struct bt_bf;

struct bt_gvc {
	unsigned char *data;

	struct bt_gvc_section files;
	struct bt_gvc_section cus;
	struct bt_gvc_section entries;
	struct bt_gvc_section reverse_lookup;

	/*
	 * Offset into the reverse lookup table where the CU basename
	 * lookup index begins.
	 */
	unsigned int cu_basename_offset;

	struct {
		char *data;
		size_t length;
	} strtab;

	struct bt_asset *asset;
	struct bt_bf **bfs;
};
typedef struct bt_gvc bt_gvc_t;

bool bt_gvc_init(bt_gvc_t *, struct bt_asset *, void *, size_t);
void bt_gvc_close(bt_gvc_t *);

/* Structs and functions for generating and writing type cache data */
struct bt_gvc_context;
typedef struct bt_gvc_context bt_gvc_context_t;

/*
 * Indicates that this context will be used to create a runtime cache for
 * immediate use.
 */
#define BT_GVC_CONTEXT_F_RUNTIME	(1UL << 0)

bt_gvc_context_t *bt_gvc_context_create(unsigned long);
void bt_gvc_context_destroy(bt_gvc_context_t *);

struct bt_so_resolver;
/*
 * Add a binary file to the global variable cache and resolve dependencies
 * using the provided resolver object.
 * If the fourth argument is non-NULL, only CUs matching names in that array
 * will be loaded.
 */
bool bt_gvc_bf(bt_gvc_context_t *, struct bt_bf *, const char *,
    struct bt_so_resolver *, BT_ARRAY(cu) *);

/* Functions for writing global variable cache data to asset file */
size_t bt_gvc_size(void *);
bool bt_gvc_write(unsigned char *, void *);

/*
 * Each resulting entry will contain pointers to relevant DWARF information
 * including bt_bf, bt_dwarf, bt_dwarf_cu and bt_dwarf_die instances.
 * Callers must confirm that this flag is set in an entry before trying to
 * access this data from that entry.
 */
#define BT_GVC_ITERATOR_DWARF		(1UL << 0)

/*
 * This flag will be set by the initialization function and should not be
 * explicitly set by the caller.
 */
#define BT_GVC_ITERATOR_BASENAME	(1UL << 1)

struct bt_gvc_iterator {
	const bt_gvc_t *cache;
	char *symbol;
	enum {
		BT_GVC_ITERATOR_NORMAL,
		BT_GVC_ITERATOR_CU,
		BT_GVC_ITERATOR_OBJECT
	} type;
	unsigned long flags;

	unsigned int index;
	unsigned int end_index;

	unsigned int chunk_index;
	unsigned int chunk_end_index;

	/*
	 * Used for filtering when both object name and CU name are specified
	 * This can be removed once indexing is improved.
	 */
	char *obj_name;
};
typedef struct bt_gvc_iterator bt_gvc_iterator_t;

struct bt_dwarf;
struct bt_dwarf_die;
struct bt_gvc_iterator_entry {
	/*
	 * Indicates which optional data was retrieved
	 */
	unsigned long flags;

	const char *symbol;
	const char *path;
	const char *cu_name;

	struct bt_bf *bf;
	struct bt_dwarf *dwarf;
	struct bt_dwarf_die *die;
	struct bt_dwarf_cu *cu;
};

bool bt_gvc_iterator_init(struct bt_gvc_iterator *, const bt_gvc_t *,
    const char *, const char *, const char *, unsigned long);
bool bt_gvc_iterator_next(struct bt_gvc_iterator *,
    struct bt_gvc_iterator_entry *);
void bt_gvc_iterator_deinit(struct bt_gvc_iterator *);

/*
 * Returns an array of names of objects that have been loaded into
 * a global variable cache.  A NULL return value indicates that an error
 * occurred.  If the return value is non-NULL, then the second argument will
 * be populated with the number of records in the array.  The array is
 * dynamically allocated and the caller is responsible for freeing it.
 * However, its contents should not be freed.
 */
const char **bt_gvc_object_list(const bt_gvc_t *, unsigned int *);

#endif /* _BT_GVC_H */
