#ifndef _BT_OS_H
#define _BT_OS_H

#include <signal.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <sys/types.h>

#include "error.h"
#include "bt_unwind.h"

#ifdef __linux__
#include <sys/ptrace.h>
#endif /* __linux__ */

/*
 * Forward declarations.
 */
struct bt_process;
struct bt_bf;
struct bt_frame;
struct bt_unwind_cursor;
struct bt_pmap;
struct bt_register_set;

/*
 * Returns any implementation-defined aliases for
 * known symbol names.
 */
const char *bt_os_pmap_alias(const char *);

/*
 * Creates PE structures for process.
 */
bool bt_os_process_maps(struct bt_process *, const char *);

/*
 * Returns an auxiliary file according to operating system.
 */
struct bt_bf *bt_os_bf_pair(struct bt_bf *, unsigned long);

/*
 * Returns an array of paths to search when resolving shared object paths.
 * The location pointed to by the argument will contain the number of paths
 * in the array.
 */
const char **bt_os_lib_paths(size_t *);

/*
 * Extract signal information from a signal frame.
 */
enum bt_process_siginfo_result bt_os_siginfo(siginfo_t *si, struct bt_frame *,
    bt_error_t *);

enum bt_pt_request {
#if defined(__linux__)
	BT_PT_GETREGS = PTRACE_GETREGS,
	BT_PT_GETFPREGS = PTRACE_GETFPREGS,
	BT_PT_DETACH = PTRACE_DETACH,
	BT_PT_ATTACH = PTRACE_ATTACH,
	BT_PT_CONTINUE = PTRACE_CONT
#elif defined(__FreeBSD__)
	BT_PT_GETREGS = PT_GETREGS,
	BT_PT_GETFPREGS = PT_GETFPREGS,
	BT_PT_DETACH = PT_DETACH,
	BT_PT_ATTACH = PT_ATTACH,
	BT_PT_CONTINUE = PT_CONTINUE
#else
	BT_PT_GETREGS = 0,
	BT_PT_GETFPREGS,
	BT_PT_DETACH,
	BT_PT_ATTACH,
	BT_PT_CONTINUE
#endif
};

#if defined(__linux__)
static inline long
bt_os_pt(enum bt_pt_request a, pid_t u, pid_t b, void *c, void *d)
{

	(void)u;

	return ptrace((enum __ptrace_request)a, b, c, d);
}
#elif defined(__sun__) || defined(__FreeBSD__)
long bt_os_pt(enum bt_pt_request, pid_t, pid_t, void *, void *);
#endif /* __sun__ || __FreeBSD__ */

/* Return the cursor's frame type */
enum bt_unwind_frame_type bt_os_unwind_frame_type(struct bt_unwind_cursor *);

/* Initializes process object for memory reads. */
bool bt_os_process_memory_init(struct bt_process *);

/* De-initializes process object for memory reads. */
void bt_os_process_memory_deinit(struct bt_process *);

/* Copies the path of the executable file into the provided buffer. */
bool bt_os_readexec(pid_t, char *, size_t, bt_error_t *);

struct bt_os_ld_link_offsets {
	/*
	 * Byte size of each map entry.
	 */
	size_t entry_size;

	/*
	 * Byte offset for the map start address.
	 */
	size_t map_offset;

	/*
	 * Byte offsets for fields within each map entry.
	 */
	size_t map_l_addr_offset;
	size_t map_name_offset;
	size_t map_ld_offset;
	size_t map_next_offset;
};

struct bt_os_ld_link_offsets *bt_os_ld_link_offsets(void);

/*
 * Some operating systems define their TLS offsets as offsets from the base
 * of the appropriate TLS block while others define them as offsets from the
 * memory map for the file.
 */
bool bt_os_tls_adjust(unsigned long long, unsigned long long *);

/*
 * Special handling of signal frames.
 */
void bt_os_sigframe(struct bt_unwind_cursor *);

/*
 * Try to unwind any OS-specific frames such as signal trampolines or
 * syscalls.  Returns BT_UNWIND_RESULT_FAILURE if no such logic applies to this
 * frame.
 */
enum bt_unwind_result bt_os_unwind(struct bt_unwind_cursor *, bt_error_t *);

#endif /* _BT_OS_H */
