#ifndef _BT_TYPE_H
#define _BT_TYPE_H

#include "dwarf.h"

#include <complex.h>
#include <stdint.h>
#include <string.h>
#include <sys/queue.h>

#define BT_TYPE_QF_NONE		(0U)
#define BT_TYPE_QF_VOLATILE	(1U << 0)
#define BT_TYPE_QF_CONST	(1U << 1)
#define BT_TYPE_QF_RESTRICT	(1U << 2)
#define BT_TYPE_QF_SHARED	(1U << 3)
#define BT_TYPE_QF_MUTABLE	(1U << 4)
#define BT_TYPE_QF_PRIVATE	(1U << 5)

enum bt_type_lang {
	BT_TYPE_LANG_C = 0,
	BT_TYPE_LANG_CPP
};

enum bt_type_kind {
	BT_TYPE_KIND_BOTTOM,
	BT_TYPE_KIND_ENUM,
	BT_TYPE_KIND_POINTER,
	BT_TYPE_KIND_ALIAS,
	BT_TYPE_KIND_TUPLE,
	BT_TYPE_KIND_ARRAY,
	BT_TYPE_KIND_BASE,
	BT_TYPE_KIND_FUNCTION,
	BT_TYPE_KIND_QUALIFIER
};

struct bt_type_encoding {
	unsigned int value;
	unsigned short n_bit;
	unsigned short n_bit_offset;
};

struct bt_type_header {
	enum bt_type_kind kind;
};

struct bt_type_key {
	bool type_unit;
	unsigned long long offset;
};

struct bt_type_enum {
	const char *label;
	long long value;
};

enum bt_type_tuple_type {
	BT_TYPE_TUPLE_STRUCT = 0,
	BT_TYPE_TUPLE_UNION,
	BT_TYPE_TUPLE_CLASS,
	BT_TYPE_TUPLE_INTERFACE
};

struct bt_type_tuple {
	unsigned int qualifiers; /* Class members. */
	const char *label;
	unsigned long long offset;
	struct bt_type *type;
	struct bt_dwarf_die *die;
};

/* Base classes of structs */
struct bt_type_parent {
	struct bt_type *type;
	struct bt_dwarf_attr_pair offset;
};

struct bt_type_arr {
	unsigned int depth;
	struct bt_type *type;
	struct bt_type_arr *dim;
	unsigned int start;
	unsigned int end;
	unsigned int block;
	struct bt_dwarf_attr_pair ub;
};

struct bt_type {
	/* The key must always be the first member. */
	struct bt_type_key key;
	struct bt_type_header h;
	unsigned int ref;
	struct bt_dwarf_die *die;
	const char *name;
	unsigned int qualifiers;

	/* Aliases and qualifiers will have base types. */
	struct bt_type *base;

	union {
		/*
		 * A NULL target and a pointer kind implies base type of
		 * void.
		 */
		struct bt_type *target;

		/*
		 * This is a value mapping for enumerators.
		 */
		struct {
			struct bt_type_enum *map;
			unsigned int length;
		} en;

		/*
		 * A value mapping for tuple array types such as structure,
		 * union and class.
		 */
		struct {
			enum bt_type_tuple_type type;

			struct {
				struct bt_type_tuple *map;
				unsigned int length;
			} fields;

			struct {
				struct bt_type_parent *map;
				unsigned int length;
			} parents;
		} tpl;

		/*
		 * A value mapping from subranges to types.
		 */
		struct bt_type_arr arr;
	} k;

	struct bt_type_encoding encoding;

	/* Pretty-printed representation of type. */
	const char *pp;

	/* Linkage for name cache. */
	struct {
		SLIST_ENTRY(bt_type) name;
	} linkage;
};

static inline enum bt_type_kind
bt_type_kind(struct bt_type *t)
{

	return t->h.kind;
}

static inline struct bt_type *
bt_type_target(struct bt_type *t)
{

	return t->k.target;
}

enum bt_variable_type {
	BT_VARIABLE_TYPE_LL = 0,	/* long long */
	BT_VARIABLE_TYPE_ULL,		/* unsigned long long */
	BT_VARIABLE_TYPE_FLT,		/* float */
	BT_VARIABLE_TYPE_DBL,		/* double */
	BT_VARIABLE_TYPE_LDBL,		/* long double */
	BT_VARIABLE_TYPE_CFLT,		/* complex float */
	BT_VARIABLE_TYPE_CDBL,		/* complex double */
	BT_VARIABLE_TYPE_CLDBL,		/* complex long double */
	BT_VARIABLE_TYPE_STRING,	/* string */
	BT_VARIABLE_TYPE_ENUM,		/* enumerator */
	BT_VARIABLE_TYPE_TUPLE,		/* tuple array */
	BT_VARIABLE_TYPE_ARRAY,		/* single-type array */
	BT_VARIABLE_TYPE_BOTTOM,	/* no type */
	BT_VARIABLE_TYPE_FUNCTION,	/* function */
	BT_VARIABLE_TYPE_REFERENCE	/* reference */
};

/*
 * bt_variable_value provides top-level value expansion.
 * A tuple consists of a sequence of variables scoped to
 * the namespace of the parent, whose values also require
 * explicit expansion through bt_variable_value.
 */
struct bt_variable_tuple {
	struct bt_variable *variable;
	unsigned int length;
	unsigned int capacity;
};

struct bt_variable_value {
	enum bt_variable_type type;
	union {
		long long ll;
		unsigned long long ull;
		float flt;
		double dbl;
		long double ldbl;
		float complex cflt;
		double complex cdbl;
		long double complex cldbl;
		struct bt_type_enum enm;
		struct bt_variable_tuple tpl;
	} v;
};
typedef struct bt_variable_value *bt_variable_value_t;

enum bt_variable_accessibility {
	BT_VARIABLE_ACCESS_NONE = 0,
	BT_VARIABLE_ACCESS_PUBLIC,
	BT_VARIABLE_ACCESS_PROTECTED,
	BT_VARIABLE_ACCESS_PRIVATE
};

static inline unsigned long long
bt_variable_value_ull(bt_variable_value_t v)
{

	return v->v.ull;
}

static inline long long
bt_variable_value_ll(bt_variable_value_t v)
{

	return v->v.ll;
}

#define BT_VARIABLE_HINT_STRING	(1UL << 0)
#define BT_VARIABLE_HINT_FPRM	(1UL << 1)

struct bt_bf;

enum bt_variable_context_type {
	BT_VARIABLE_CONTEXT_FRAME,	/* Stack variables */
	BT_VARIABLE_CONTEXT_THREAD,	/* TLS variables */
	BT_VARIABLE_CONTEXT_PROCESS	/* Global variables*/
};

struct bt_variable_context {
	enum bt_variable_context_type type;
	union {
		struct bt_frame *frame;
		struct bt_thread *thread;
		struct bt_process *process;
	} context;

	/* Only valid for thread/process-scoped variables */
	const char *obj_name;
	const char *cu_name;
};

#define BT_VARIABLE_S_NONE	0
#define BT_VARIABLE_S_FAILED	(1UL << 0)
#define BT_VARIABLE_S_SUCCESS	(1UL << 1)

struct bt_variable {
	/* The variable may be a member of another variable. */
	struct bt_variable *parent;

	/*
	 * If this is a member variable inherited from a base class then we
	 * cache any additional offset within the child class.
	 */
	unsigned int inheritance_offset;

	/* Argument offset. */
	unsigned int fprm_offset;

	/* A variable may have additional type qualifiers. */
	unsigned int qualifiers;

	/* Formatting hints. */
	unsigned long hint;

	/* Array offset. */
	unsigned long long base;
	unsigned long long offset;

	/* Flag to indicate whether value has been resolved. */
	unsigned long state;

	/*
	 * Indicates whether this variable is thread-local.  This field is
	 * only valid if the variable has been resolved.
	 */
	bool tls;

	unsigned int bit_offset;
	unsigned int bit_size;
	enum bt_variable_accessibility accessibility;
	struct bt_type *type;
	const char *name;
	unsigned long long address;
	unsigned long long reference;
	struct {
		unsigned long flags;
		struct bt_dwarf_attr_pair value;
	} location;
	struct bt_variable_context context;
	struct bt_dwarf_die *die;
	struct bt_dwarf *dwarf;
	unsigned long long c_base;
	struct bt_variable_value value;

	/*
	 * Debug information is not always reliable so we may get a specific
	 * location for a variable from alternate sources such as the symbol
	 * table.  If this field is non-zero, then it should be used as the
	 * variable's location instead of the debug information.
	 */
	unsigned long long override_address;

	/* Error state associated with variable. */
	bt_error_t error;

	STAILQ_ENTRY(bt_variable) linkage;
};
typedef struct bt_variable bt_variable_t;

struct bt_variable_array {
	bt_variable_t **values;
	size_t size;
};
typedef struct bt_variable_array bt_variable_array_t;

static inline const char *
bt_variable_error(const struct bt_variable *v)
{

	return bt_error_message(&v->error);
}

static inline const char *
bt_variable_name(const struct bt_variable *v)
{

	return v->name;
}

static inline enum bt_variable_accessibility
bt_variable_accessibility(const struct bt_variable *v)
{

	return v->accessibility;
}

static inline enum bt_variable_context_type
bt_variable_context_type(const struct bt_variable *v)
{

	return v->context.type;
}

static inline size_t
bt_variable_ptr_size(const struct bt_variable *v)
{

	return bt_dwarf_ptr_size(v->dwarf);
}

static inline bool
bt_variable_resolved(const struct bt_variable *v)
{

	return v->state & BT_VARIABLE_S_SUCCESS;
}

static inline bt_dwarf_t *
bt_variable_dwarf(const struct bt_variable *v)
{

	return v->dwarf;
}

/*
 * Return the appropriate cursor for the variable's context.  This cursor
 * should be used for any evaluation regarding this variable.
 */
struct bt_unwind_cursor *bt_variable_cursor(struct bt_variable *);

/*
 * Returns the name of the variable's CU.  This will be NULL for non-global
 * variables.
 */
const char *bt_variable_cu_name(const struct bt_variable *);

/*
 * Returns the name of the variable's object file.  This will be NULL for
 * non-global variables.
 */
const char *bt_variable_obj_name(const struct bt_variable *);

/*
 * Returns the process handle for a variable.  Returns NULL for frame variables.
 */
struct bt_process *bt_variable_process(struct bt_variable *);

/*
 * Returns the thread handle for a variable.  Only valid for variables with
 * thread context.
 */
struct bt_thread *bt_variable_thread(struct bt_variable *);

struct bt_variable_array_iterator {
	unsigned int end;
	struct bt_variable *array;
	struct bt_type_arr *arr;
	unsigned int offset;
	unsigned int base;
};
typedef struct bt_variable_array_iterator bt_variable_array_iterator_t;

bool bt_variable_array_iterator_init(bt_variable_array_iterator_t *,
    bt_variable_t *);
bool bt_variable_array_iterator_nest(bt_variable_array_iterator_t *,
    bt_variable_array_iterator_t *);

enum bt_variable_array_action {
	BT_VARIABLE_ARRAY_END,
	BT_VARIABLE_ARRAY_VALUE,
	BT_VARIABLE_ARRAY_NEST
};

enum bt_variable_array_action
    bt_variable_array_iterate(bt_variable_array_iterator_t *,
    bt_variable_t *, size_t *);

/*
 * Returns true if the specified variable is an array, and sets the
 * second argument to the length of the array.
 */
bool bt_variable_array_length(bt_variable_t *, size_t *);

/*
 * Synthesizes a variable object from a reference-type variable.
 */
bool bt_variable_deref(bt_variable_t *, bt_variable_t *, bt_error_t *);

/*
 * Synthesizes a variable using derived address.
 */
bool bt_variable_deref_at(bt_variable_t *, bt_variable_t *,
    unsigned long long, bt_error_t *);

/*
 * Synthesizes a variable of the given type using the given address.
 */
bool bt_variable_type_deref_at(bt_variable_t *, bt_variable_t *,
    struct bt_type *, unsigned long long);

/*
 * Cast a variable from one type (second argument) to another (first argument).
 */
void bt_variable_cast(struct bt_variable *output, struct bt_variable *input,
    struct bt_type *type);

static inline bt_variable_t *
bt_variable_parent(struct bt_variable *v)
{

	return v->parent;
}

static inline unsigned long
bt_variable_hint(struct bt_variable *v)
{

	return v->hint;
}

struct bt_type *bt_type_base(struct bt_type *);
size_t bt_type_sizeof(struct bt_type *);
struct bt_type *bt_type_deref(struct bt_type *);
ssize_t bt_variable_strncpy(char *, bt_variable_t *, size_t);
bool bt_variable_value(bt_variable_t *);
long long bt_variable_enum(bt_variable_t *, const char **);
struct bt_type *bt_type_bottom(void);
const char *bt_type_pp(struct bt_type *, enum bt_type_lang);
void bt_variable_deinit(bt_variable_t *);
void bt_variable_destroy(bt_variable_t *);

/*
 * Returns true if the character specified in the third argument
 * is found in the first N characters of the string. The index
 * of the character is stored in the location pointed to by the
 * last argument.
 */
bool bt_variable_strnchr(bt_variable_t *, size_t, int, size_t *);

/*
 * This must be called after value expansion.
 */
size_t bt_variable_sizeof(const bt_variable_t *);

static inline long long int
bt_variable_fprm_offset(struct bt_variable *v)
{

	if (bt_variable_hint(v) & BT_VARIABLE_HINT_FPRM)
		return v->fprm_offset - 1;

	return -1;
}

static inline enum bt_variable_type
bt_variable_type(struct bt_variable *v)
{

	return v->value.type;
}

static inline struct bt_type *
bt_variable_bt_type(struct bt_variable *v)
{

	return v->type;
}

static inline unsigned long long
bt_variable_address(struct bt_variable *v)
{

	return v->address;
}

typedef unsigned long long bt_variable_ref_t;

/*
 * The first variable is set to point to variable[size]. This currently
 * only works for variables in memory.
 */
bool bt_variable_xderef(bt_variable_t *, bt_variable_t *, size_t,
    bt_error_t *);

/*
 * Same as bt_variable_xderef except it applies the mask specified in the
 * last argument to the resolved address.
 */
bool bt_variable_xderef_mask(bt_variable_t *, bt_variable_t *, size_t,
    unsigned long long, bt_error_t *);

#define BT_VARIABLE_FOREACH_MEMBER_BODY(M, V, N)		\
	for (unsigned int (_bt_i_##N) = 0;			\
	    (_bt_i_##N) < (V)->value.v.tpl.length &&		\
	    (((M) = &(V)->value.v.tpl.variable[_bt_i_##N]), 1);	\
	    (_bt_i_##N)++)

#define BT_VARIABLE_FOREACH_MEMBER_STUB(M, V, N)	\
	BT_VARIABLE_FOREACH_MEMBER_BODY(M, V, N)

#define BT_VARIABLE_FOREACH_MEMBER(M, V)		\
	BT_VARIABLE_FOREACH_MEMBER_STUB(M, V, __COUNTER__)

static inline bt_variable_t *
bt_variable_member_at(struct bt_variable *v, unsigned int i)
{

	if (i >= v->value.v.tpl.length)
		return NULL;

	return &v->value.v.tpl.variable[i];
}

/*
 * Returns a pointer to uninitialized member variable of the passed object.
 * bt_variable_value() needs to be called on the resultant pointer to access
 * its value.
 * Returns NULL on failure.
 */
bt_variable_t *bt_variable_member_get(bt_variable_t *, const char *);

/*
 * Returns a pointer to initialized member variable of the passed object.
 * Returns NULL on failure.
 */
bt_variable_t *bt_variable_member_value_get(bt_variable_t *, const char *);

/*
 * Returns an array containing pointers to all members that have matching name.
 */
bt_variable_array_t *bt_variable_all_members_get(bt_variable_t *, const char *);

/*
 * Frees the array object.
 */
void bt_variable_array_free(bt_variable_array_t *);

#define BT_VARIABLE_P(S, O, F)				\
	static inline O					\
	bt_variable_##S(struct bt_variable *v)		\
	{						\
		return v->value.v.F;			\
	}
#define BT_VARIABLE_PTR_PRI "#llx"
BT_VARIABLE_P(ref, bt_variable_ref_t, ull);
BT_VARIABLE_P(cdbl, double complex, cdbl);
BT_VARIABLE_P(cldbl, long double complex, cldbl);
BT_VARIABLE_P(cflt, float complex, cflt);
BT_VARIABLE_P(dbl, double, dbl);
BT_VARIABLE_P(ldbl, long double, ldbl);
BT_VARIABLE_P(flt, float, flt);
BT_VARIABLE_P(ull, unsigned long long, ull);
BT_VARIABLE_P(ll, long long, ll);
#undef BT_VARIABLE_P

static inline bool
bt_variable_int(int *ptr, bt_variable_t *var, bt_error_t *e)
{
	long long v = bt_variable_ll(var);

	if (v >= INT_MIN && v <= INT_MAX) {
		*ptr = (int)v;
		return true;
	}
	return bt_error_set(e, "value out of range");
}

static inline int
bt_variable_enumint(int *ptr, struct bt_variable *var, bt_error_t *e)
{
	const char *e_l;
	long long v = bt_variable_enum(var, &e_l);

	if (v >= INT_MIN && v <= INT_MAX) {
		*ptr = (int)v;
		return true;
	}
	return bt_error_set(e, "value out of range");
}

#endif /* _BT_TYPE_H */
